# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.mport threading


from unittest import mock

import oslotest.base

import designate.conf
from designate import rpc
import designate.tests
from designate.worker import rpcapi


CONF = designate.conf.CONF


class TestService(oslotest.base.BaseTestCase):
    @mock.patch.object(rpc, 'get_client', mock.Mock())
    def setUp(self):
        super().setUp()

        self.worker = rpcapi.WorkerAPI()
        self.worker.client = mock.Mock()

    def test_create_zone(self):
        self.worker.create_zone('context', 'zone')

        self.worker.client.cast.assert_called_with(
            'context', 'create_zone', zone='zone'
        )

    def test_update_zone(self):
        self.worker.update_zone('context', 'zone')

        self.worker.client.cast.assert_called_with(
            'context', 'update_zone', zone='zone'
        )

    def test_delete_zone(self):
        self.worker.delete_zone('context', 'zone', 'hard_delete')

        self.worker.client.cast.assert_called_with(
            'context', 'delete_zone', zone='zone', hard_delete='hard_delete'
        )

    def test_recover_shard(self):
        self.worker.recover_shard('context', 'begin', 'end')

        self.worker.client.cast.assert_called_with(
            'context', 'recover_shard', begin='begin', end='end'
        )

    def test_start_zone_export(self):
        self.worker.start_zone_export('context', 'zone', 'export')

        self.worker.client.cast.assert_called_with(
            'context', 'start_zone_export', zone='zone', export='export'
        )

    def test_perform_zone_xfr(self):
        self.worker.perform_zone_xfr('context', 'zone', 'servers')

        self.worker.client.cast.assert_called_with(
            'context', 'perform_zone_xfr', zone='zone', servers='servers'
        )

    def test_get_serial_number(self):
        self.worker.get_serial_number('context', 'zone', 'host', 53)

        self.worker.client.call.assert_called_with(
            'context', 'get_serial_number', zone='zone', host='host', port=53
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import oslotest.base

from designate.common.decorators import lock
from designate import objects


class TestExtractZoneId(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()

    def test_extract_zone_id_empty(self):
        self.assertIsNone(lock.extract_zone_id([], {}))

    def test_extract_zone_id_no_valid_objects(self):
        self.assertIsNone(
            lock.extract_zone_id([], {
                'ptr': objects.PTRList(), 'a': objects.AList()})
        )

    def test_extract_zone_id_kwargs(self):
        self.assertEqual(
            'test',
            lock.extract_zone_id([], {'zone_id': 'test'})
        )
        self.assertEqual(
            'test',
            lock.extract_zone_id([], {'zone': mock.Mock(id='test')})
        )
        self.assertEqual(
            'test',
            lock.extract_zone_id([], {'recordset': mock.Mock(zone_id='test')})
        )
        self.assertEqual(
            'test',
            lock.extract_zone_id([], {'record': mock.Mock(zone_id='test')})
        )

    def test_extract_zone_id_from_zone(self):
        self.assertEqual(
            '123',
            lock.extract_zone_id(['a', 'b', 'c'], {'x': objects.Zone(id=123)})
        )
        self.assertEqual(
            '123',
            lock.extract_zone_id([objects.Zone(id=123)], {})
        )

    def test_extract_zone_id_from_recordset(self):
        self.assertEqual(
            '123',
            lock.extract_zone_id([], {'x': objects.RecordSet(zone_id=123)})
        )
        self.assertEqual(
            '123',
            lock.extract_zone_id([objects.RecordSet(zone_id=123)], {})
        )

    def test_extract_zone_id_from_record(self):
        self.assertEqual(
            '123',
            lock.extract_zone_id([], {'x': objects.Record(zone_id=123)})
        )
        self.assertEqual(
            '123',
            lock.extract_zone_id([objects.Record(zone_id=123)], {})
        )

    def test_extract_zone_id_from_zone_transfer_request(self):
        self.assertEqual(
            '123',
            lock.extract_zone_id(
                [], {'x': objects.ZoneTransferRequest(zone_id=123)})
        )
        self.assertEqual(
            '123',
            lock.extract_zone_id(
                [objects.ZoneTransferRequest(zone_id=123)], {})
        )

    def test_extract_zone_id_from_zone_transfer_accept(self):
        self.assertEqual(
            '123',
            lock.extract_zone_id(
                [], {'x': objects.ZoneTransferAccept(zone_id=123)})
        )
        self.assertEqual(
            '123',
            lock.extract_zone_id([objects.ZoneTransferAccept(zone_id=123)], {})
        )

    def test_extract_zone_id_from_second_argument(self):
        self.assertEqual('456', lock.extract_zone_id(['123', '456'], {}))

    def test_extract_zone_id_when_second_argument_is_a_zone(self):
        self.assertEqual(
            '456', lock.extract_zone_id(['123', objects.Zone(id=456)], {})
        )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

import oslotest.base

import designate.conf
from designate import dnsmiddleware
from designate.mdns import service
from designate import policy
from designate import rpc
import designate.service
from designate import storage
from designate.tests import base_fixtures
import designate.utils


CONF = designate.conf.CONF


class MdnsServiceTest(oslotest.base.BaseTestCase):
    @mock.patch.object(storage, 'get_storage', mock.Mock())
    @mock.patch.object(policy, 'init')
    @mock.patch.object(rpc, 'init')
    @mock.patch.object(rpc, 'initialized')
    def setUp(self, mock_rpc_initialized, mock_rpc_init, mock_policy_init):
        super().setUp()
        self.stdlog = base_fixtures.StandardLogging()
        self.useFixture(self.stdlog)

        mock_rpc_initialized.return_value = False

        self.service = service.Service()

        mock_policy_init.assert_called_once()
        mock_rpc_initialized.assert_called_once()
        mock_rpc_init.assert_called_once()

    @mock.patch.object(designate.service.DNSService, 'start')
    def test_service_start(self, mock_dns_start):
        self.service.start()

        mock_dns_start.assert_called()

    def test_service_stop(self):
        self.service.dns_service.stop = mock.Mock()

        self.service.stop()

        self.service.dns_service.stop.assert_called()

        self.assertIn('Stopping mdns service', self.stdlog.logger.output)

    def test_service_name(self):
        self.assertEqual('mdns', self.service.service_name)

    def test_dns_application(self):
        app = self.service.dns_application
        self.assertIsInstance(app, dnsmiddleware.DNSMiddleware)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import oslotest.base

from designate.objects import fields


class EnumTest(oslotest.base.BaseTestCase):
    def test_get_schema(self):
        result = fields.Enum('valid').get_schema()
        self.assertEqual('valid', result['enum'])
        self.assertEqual('any', result['type'])


class DomainFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_domain_field(self):
        sub_domain = 'a' * 64
        self.assertRaisesRegex(
            ValueError,
            f'Host {sub_domain} is too long',
            fields.DomainField().coerce, mock.Mock(), mock.Mock(),
            f'{sub_domain}.example.org.'
        )


class EmailFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_email_field(self):
        self.assertRaisesRegex(
            ValueError,
            '- is not an email',
            fields.EmailField().coerce, mock.Mock(), mock.Mock(),
            '-'
        )
        self.assertRaisesRegex(
            ValueError,
            'Email @ is invalid',
            fields.EmailField().coerce, mock.Mock(), mock.Mock(),
            '@'
        )


class HostFieldTest(oslotest.base.BaseTestCase):
    def test_is_none(self):
        self.assertIsNone(
            fields.HostField(nullable=True).coerce(
                mock.Mock(), mock.Mock(), None
            )
        )

    def test_not_valid_host_field(self):
        sub_domain = 'a' * 64
        self.assertRaisesRegex(
            ValueError,
            f'Host {sub_domain} is too long',
            fields.HostField().coerce, mock.Mock(), mock.Mock(),
            f'{sub_domain}.example.org.'
        )
        self.assertRaisesRegex(
            ValueError,
            'Host name example.org does not end with a dot',
            fields.HostField().coerce, mock.Mock(), mock.Mock(),
            'example.org'
        )


class SRVFieldTest(oslotest.base.BaseTestCase):
    def test_is_none(self):
        self.assertIsNone(
            fields.SRVField(nullable=True).coerce(
                mock.Mock(), mock.Mock(), None
            )
        )

    def test_not_valid_srv_field(self):
        sub_domain = 'a' * 64
        self.assertRaisesRegex(
            ValueError,
            f'Host {sub_domain} is too long',
            fields.SRVField().coerce, mock.Mock(), mock.Mock(),
            f'{sub_domain}.example.org.'
        )
        self.assertRaisesRegex(
            ValueError,
            'Host name example.org does not end with a dot',
            fields.SRVField().coerce, mock.Mock(), mock.Mock(),
            'example.org'
        )


class TxtFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_txt(self):
        self.assertRaisesRegex(
            ValueError,
            r"Do NOT put '\\' into end of TXT record",
            fields.TxtField().coerce, mock.Mock(), mock.Mock(),
            " \\"
        )


class SshfpTest(oslotest.base.BaseTestCase):
    def test_not_valid_sshfp(self):
        self.assertRaisesRegex(
            ValueError,
            'Host name  is not a SSHFP record',
            fields.Sshfp().coerce, mock.Mock(), mock.Mock(),
            ''
        )


class NaptrFlagsFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_naptr_flags(self):
        record = 'A' * 256
        self.assertRaisesRegex(
            ValueError,
            f'NAPTR record {record} flags field cannot be longer than 255 '
            'characters',
            fields.NaptrFlagsField().coerce, mock.Mock(), mock.Mock(),
            record
        )
        self.assertRaisesRegex(
            ValueError,
            'NAPTR record record flags can be S, A, U and P',
            fields.NaptrFlagsField().coerce, mock.Mock(), mock.Mock(),
            'record'
        )


class NaptrServiceFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_naptr_service(self):
        record = 'A' * 256
        self.assertRaisesRegex(
            ValueError,
            f'NAPTR record {record} service field cannot be longer than 255 '
            'characters',
            fields.NaptrServiceField().coerce, mock.Mock(), mock.Mock(),
            record
        )
        self.assertRaisesRegex(
            ValueError,
            '_ NAPTR record service is invalid',
            fields.NaptrServiceField().coerce, mock.Mock(), mock.Mock(),
            '_'
        )


class NaptrRegexpFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_naptr_regexp(self):
        record = 'A' * 256
        self.assertRaisesRegex(
            ValueError,
            f'NAPTR record {record} regexp field cannot be longer than 255 '
            'characters',
            fields.NaptrRegexpField().coerce, mock.Mock(), mock.Mock(),
            record
        )
        self.assertRaisesRegex(
            ValueError,
            '_ NAPTR record is invalid',
            fields.NaptrRegexpField().coerce, mock.Mock(), mock.Mock(),
            '_'
        )


class CertTypeFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_cert_type(self):
        self.assertRaisesRegex(
            ValueError,
            'Cert type _ is not a valid Mnemonic or value',
            fields.CertTypeField().coerce, mock.Mock(), mock.Mock(),
            '_'
        )


class CertAlgoFieldTest(oslotest.base.BaseTestCase):
    def test_not_valid_cert_algo(self):
        self.assertRaisesRegex(
            ValueError,
            'Cert Algo _ is not a valid Mnemonic or value',
            fields.CertAlgoField().coerce, mock.Mock(), mock.Mock(),
            '_'
        )


class IPOrHostTest(oslotest.base.BaseTestCase):
    def test_valid(self):
        self.assertEqual(
            'example.org.',
            fields.IPOrHost().coerce(mock.Mock(), mock.Mock(), 'example.org.')
        )
        self.assertEqual(
            '192.0.2.1',
            fields.IPOrHost().coerce(mock.Mock(), mock.Mock(), '192.0.2.1')
        )

    def test_not_valid_ip_or_host(self):
        self.assertRaisesRegex(
            ValueError,
            'example.org is not IP address or host name',
            fields.IPOrHost().coerce, mock.Mock(), mock.Mock(), 'example.org'
        )



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate import exceptions
from designate import objects
from designate.objects.adapters.api_v2 import base


class RecordSetAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.RecordSet
    MODIFICATIONS = {
        'fields': {
            "id": {},
            "zone_id": {},
            "project_id": {
                'rename': 'tenant_id'
            },
            "name": {
                'immutable': True
            },
            "zone_name": {
                'read_only': True,
            },
            "type": {
                'rename': 'type',
                'immutable': True
            },
            "records": {
                'read_only': False
            },
            "description": {
                'read_only': False
            },
            "ttl": {
                'read_only': False
            },
            "status": {},
            "action": {},
            "version": {},
            "created_at": {},
            "updated_at": {},
        },
        'options': {
            'links': True,
            'resource_name': 'recordset',
            'collection_name': 'recordsets',
        }
    }

    @classmethod
    def parse_object(cls, new_recordset, recordset, *args, **kwargs):
        # TODO(Graham): Remove this when
        # https://bugs.launchpad.net/designate/+bug/1432842 is fixed
        try:
            recordset.records
        except exceptions.RelationNotLoaded:
            recordset.records = objects.RecordList()

        original_records = set()
        for record in recordset.records:
            original_records.add(record.data)
        # Get new list of Records
        new_records = set()
        if 'records' in new_recordset:
            if isinstance(new_recordset['records'], list):
                for record in new_recordset['records']:
                    new_records.add(record)
            else:
                errors = objects.ValidationErrorList()
                e = objects.ValidationError()
                e.path = ['records']
                e.validator = 'type'
                e.validator_value = ["list"]
                e.message = ("'%(data)s' is not a valid list of records"
                             % {'data': new_recordset['records']})
                # Add it to the list for later
                errors.append(e)
                raise exceptions.InvalidObject(
                    "Provided object does not match "
                    "schema", errors=errors, object=cls.ADAPTER_OBJECT())

        # Get differences of Records
        records_to_add = new_records.difference(original_records)
        records_to_rm = original_records.difference(new_records)

        # Update all items except records
        record_update = False
        if 'records' in new_recordset:
            record_update = True
            del new_recordset['records']

        if record_update:
            # Build a list of the "new" records
            new_recordset_records = objects.RecordList()

            # Remove deleted records if we have provided a records array
            for record in recordset.records:
                if record.data not in records_to_rm:
                    new_recordset_records.append(record)

            # Add new records
            for record in records_to_add:
                new_recordset_records.append(objects.Record(data=record))

            # Do a single assignment, preserves the object change fields
            recordset.records = new_recordset_records

        return super().parse_object(
            new_recordset, recordset, *args, **kwargs)

    @classmethod
    def _get_path(cls, request, obj):
        ori_path = request.path
        path = ori_path.lstrip('/').split('/')
        insert_zones = False
        to_insert = ''
        if 'zones' not in path and obj is not None:
            insert_zones = True
            to_insert = f'zones/{obj.zone_id}'

        item_path = ''
        for part in path:
            if part == cls.MODIFICATIONS['options']['collection_name']:
                item_path += '/' + part
                return item_path
            elif insert_zones and to_insert and part == 'v2':
                item_path += f'/v2/{to_insert}'
                insert_zones = False  # make sure only insert once if needed
            else:
                item_path += '/' + part


class RecordSetListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.RecordSetList
    MODIFICATIONS = {
        'options': {
            'links': True,
            'resource_name': 'recordset',
            'collection_name': 'recordsets',
        }
    }



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime
from unittest import mock

from oslo_log import log as logging
from oslo_utils import timeutils
import oslotest.base

from designate import exceptions
from designate import objects
from designate.objects import adapters
from designate.objects import base
from designate.objects import fields

LOG = logging.getLogger(__name__)


@base.DesignateRegistry.register
class DesignateTestObject(base.DictObjectMixin, base.PersistentObjectMixin,
                          base.DesignateObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    fields = {
        'name': fields.StringFields(maxLength=255),
        'description': fields.StringFields(nullable=True, maxLength=255)
    }

    STRING_KEYS = [
        'id', 'name'
    ]


@base.DesignateRegistry.register
class DesignateTestPersistentObject(objects.base.PersistentObjectMixin,
                                    objects.DesignateObject):
    pass


class DesignateTestAdapter(adapters.DesignateAdapter):
    ADAPTER_OBJECT = DesignateTestObject
    ADAPTER_FORMAT = 'TEST_API'

    MODIFICATIONS = {
        'fields': {
            'id': {},
            'name': {
                'read_only': False
            },
            'description': {
                'read_only': False
            },
            'created_at': {},
            'updated_at': {},
        },
        'options': {
            'links': True,
            'resource_name': 'test_obj',
            'collection_name': 'test_obj',
        }
    }


class DesignateDateTimeAdaptor(adapters.DesignateAdapter):
    ADAPTER_OBJECT = DesignateTestPersistentObject
    ADAPTER_FORMAT = 'TEST_API'

    MODIFICATIONS = {
        'fields': {
            'id': {},
            'created_at': {},
            'updated_at': {},
        },
        'options': {}
    }


class DesignateAdapterTest(oslotest.base.BaseTestCase):
    def test_parse(self):
        test_obj = adapters.DesignateAdapter.parse(
            'TEST_API', {'name': 'example.test.'}, DesignateTestObject()
        )

        self.assertIsInstance(test_obj, DesignateTestObject)
        self.assertEqual('example.test.', test_obj.name)

    @mock.patch.object(DesignateTestAdapter, 'parse_object')
    def test_parse_type_error(self, mock_parse_object):
        mock_parse_object.side_effect = TypeError('test')

        self.assertRaisesRegex(
            exceptions.InvalidObject,
            'Provided object is not valid. Got a TypeError with message test',
            adapters.DesignateAdapter.parse, 'TEST_API', {},
            DesignateTestObject()
        )

    @mock.patch.object(DesignateTestAdapter, 'parse_object')
    def test_parse_attribute_error(self, mock_parse_object):
        mock_parse_object.side_effect = AttributeError('test')

        self.assertRaisesRegex(
            exceptions.InvalidObject,
            'Provided object is not valid. Got an AttributeError '
            'with message test',
            adapters.DesignateAdapter.parse, 'TEST_API', {},
            DesignateTestObject()
        )

    def test_parse_schema_does_not_match(self):
        self.assertRaisesRegex(
            exceptions.InvalidObject,
            'Provided object does not match schema. '
            'Keys \\[\'address\'\\] are not valid for test_obj',
            adapters.DesignateAdapter.parse,
            'TEST_API', {'address': '192.0.2.1'}, DesignateTestObject(),
        )

    def test_get_object_adapter(self):
        adapter = adapters.DesignateAdapter.get_object_adapter(
            DesignateTestObject(), 'TEST_API'
        )

        self.assertIsInstance(adapter(), DesignateTestAdapter)

    def test_get_object_adapter_not_found(self):
        self.assertRaisesRegex(
            exceptions.AdapterNotFound,
            'Adapter for DesignateTestObject to format None not found',
            adapters.DesignateAdapter.get_object_adapter, DesignateTestObject()
        )

    def test_object_render(self):
        test_obj = adapters.DesignateAdapter.render(
            'TEST_API', DesignateTestObject()
        )

        self.assertEqual(
            sorted([
                'created_at', 'description', 'id', 'name', 'updated_at',
            ]),
            sorted(test_obj)
        )

    def test_datetime_format(self):
        now = timeutils.utcnow()
        test_obj = DesignateTestPersistentObject()
        test_obj.created_at = now
        test_dict = adapters.DesignateAdapter.render('TEST_API', test_obj)

        self.assertEqual(
            datetime.datetime.strptime(
                test_dict['created_at'], '%Y-%m-%dT%H:%M:%S.%f'
            ),
            test_obj.created_at
        )



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import random
import socket

import dns.exception
import dns.message
import dns.opcode
import dns.query
import dns.rdatatype
import dns.zone
import eventlet
from oslo_log import log as logging
from oslo_serialization import base64

import designate.conf
from designate import context
from designate import exceptions
from designate import objects

CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class TsigKeyring(dict):
    """Implements the DNSPython KeyRing API, backed by the Designate DB"""

    def __init__(self, storage):
        super().__init__()
        self.storage = storage

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key, default=None):
        try:
            name = key.to_text(True)
            criterion = {'name': name}
            tsigkey = self.storage.find_tsigkey(
                context.get_current(), criterion
            )

            return base64.decode_as_bytes(tsigkey.secret)

        except exceptions.TsigKeyNotFound:
            return default


def from_dnspython_zone(dnspython_zone):
    # dnspython never builds a zone with more than one SOA, even if we give
    # it a zonefile that contains more than one
    soa = dnspython_zone.get_rdataset(dnspython_zone.origin, 'SOA')
    if soa is None:
        raise exceptions.BadRequest('An SOA record is required')
    if soa.ttl == 0:
        soa.ttl = CONF['service:central'].min_ttl
    email = soa[0].rname.to_text(omit_final_dot=True)
    email = email.replace('.', '@', 1)

    name = dnspython_zone.origin.to_text()
    values = {
        'name': name,
        'email': email,
        'ttl': soa.ttl,
        'serial': soa[0].serial,
        'retry': soa[0].retry,
        'expire': soa[0].expire
    }

    zone = objects.Zone(**values)

    rrsets = dnspyrecords_to_recordsetlist(dnspython_zone.nodes)
    zone.recordsets = rrsets
    return zone


def dnspyrecords_to_recordsetlist(dnspython_records):
    rrsets = objects.RecordSetList()
    for rname in dnspython_records.keys():
        for rdataset in dnspython_records[rname]:
            rrsets.append(dnspythonrecord_to_recordset(rname, rdataset))
    return rrsets


def dnspythonrecord_to_recordset(rname, rdataset):
    record_type = dns.rdatatype.to_text(rdataset.rdtype)

    name = rname.to_text()

    # Create the other recordsets

    values = {
        'name': name,
        'type': record_type
    }

    if rdataset.ttl != 0:
        values['ttl'] = rdataset.ttl

    rrset = objects.RecordSet(**values)
    rrset.records = objects.RecordList()

    for rdata in rdataset:
        rr = objects.Record(data=rdata.to_text())
        rrset.records.append(rr)
    return rrset


def xfr_timeout():
    return CONF['service:worker'].xfr_timeout


def do_axfr(zone_name, servers, source=None):
    """
    Requests an AXFR for a given zone name and process the response

    :returns: Zone instance from dnspython
    """
    random.shuffle(servers)

    xfr = None
    for srv in servers:
        for address in get_ip_addresses(srv['host']):
            to = eventlet.Timeout(xfr_timeout())
            log_info = {'name': zone_name, 'host': srv, 'address': address}
            try:
                LOG.info(
                    'Doing AXFR for %(name)s from %(host)s %(address)s',
                    log_info
                )
                xfr = dns.query.xfr(
                    address, zone_name, relativize=False, timeout=1,
                    port=srv['port'], source=source
                )
                raw_zone = dns.zone.from_xfr(xfr, relativize=False)
                LOG.debug('AXFR Successful for %s', raw_zone.origin.to_text())
                return raw_zone
            except eventlet.Timeout as t:
                if t == to:
                    LOG.error('AXFR timed out for %(name)s from %(host)s',
                              log_info)
                    continue
            except dns.exception.FormError:
                LOG.error('Zone %(name)s is not present on %(host)s.'
                          'Trying next server.', log_info)
            except OSError:
                LOG.error('Connection error when doing AXFR for %(name)s from '
                          '%(host)s', log_info)
            except Exception:
                LOG.exception('Problem doing AXFR %(name)s from %(host)s. '
                              'Trying next server.', log_info)
            finally:
                to.cancel()

    raise exceptions.XFRFailure(
        'XFR failed for %(name)s. No servers in %(servers)s was reached.' %
        {'name': zone_name, 'servers': servers}
    )


def prepare_dns_message(zone_name, rdatatype, opcode):
    """
    Create a dns message using dnspython
    """
    dns_message = dns.message.make_query(zone_name, rdatatype)
    dns_message.set_opcode(opcode)
    return dns_message


def notify(zone_name, host, port=53, timeout=10):
    """
    Create a NOTIFY message and send it
    """
    dns_message = prepare_dns_message(
        zone_name, rdatatype=dns.rdatatype.SOA, opcode=dns.opcode.NOTIFY
    )
    return send_dns_message(dns_message, host, port=port, timeout=timeout)


def soa_query(zone_name, host, port=53, timeout=10):
    """
    Create a SOA Query message and send it
    """
    dns_message = prepare_dns_message(
        zone_name, rdatatype=dns.rdatatype.SOA, opcode=dns.opcode.QUERY
    )
    return send_dns_message(dns_message, host, port=port, timeout=timeout)


def use_all_tcp():
    return CONF['service:worker'].all_tcp


def send_dns_message(dns_message, host, port=53, timeout=10):
    """
    Send the dns message and return the response

    :return: dns.Message of the response to the dns query
    """
    ip_address = get_ip_address(host)
    # This can raise some exceptions, but we'll catch them elsewhere
    if not use_all_tcp():
        return dns.query.udp(
            dns_message, ip_address, port=port, timeout=timeout)
    return dns.query.tcp(
        dns_message, ip_address, port=port, timeout=timeout)


def get_serial(zone_name, host, port=53):
    """
    Possibly raises dns.exception.Timeout or dns.query.BadResponse.
    Possibly returns 0 if, e.g., the answer section is empty.
    """
    resp = soa_query(zone_name, host, port=port)
    if not resp.answer:
        return 0
    rdataset = resp.answer[0].to_rdataset()
    if not rdataset:
        return 0
    return rdataset[0].serial


def get_ip_address(ip_address_or_hostname):
    """
    Provide an ip or hostname and return a valid ip4 or ipv6 address.

    :return: ip address
    """
    addresses = get_ip_addresses(ip_address_or_hostname)
    if not addresses:
        return None
    return addresses[0]


def get_ip_addresses(ip_address_or_hostname):
    """
    Provide an ip or hostname and return all valid ip4 or ipv6 addresses.

    :return: ip addresses
    """
    addresses = []
    for res in socket.getaddrinfo(ip_address_or_hostname, 0):
        addresses.append(res[4][0])
    return list(set(addresses))



# Copyright 2022 Cloudification GmbH
#
# Author: Kiran P <contact@cloudification.io>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
import pecan

from designate.api.v2.controllers import rest
from designate import exceptions
from designate.objects.adapters import DesignateAdapter
from designate import utils

LOG = logging.getLogger(__name__)


class PoolMoveController(rest.RestController):

    @pecan.expose(template='json:', content_type='application/json')
    @pecan.expose(template='json:', content_type='application/json-patch+json')
    @utils.validate_uuid('zone_id')
    def post_all(self, zone_id):
        """Move a zone"""
        request = pecan.request
        response = pecan.response
        body = request.body_dict
        context = request.environ['context']

        zone = self.central_api.get_zone(context, zone_id)

        if zone.action == "DELETE":
            raise exceptions.BadRequest('Can not move a deleting zone')

        target_pool_id = None
        if 'pool_id' in body:
            if zone.pool_id == body['pool_id']:
                raise exceptions.BadRequest(
                    'Target pool must be different for zone pool move'
                )
            target_pool_id = body['pool_id']

        # Update the zone object with the new values
        zone = DesignateAdapter.parse('API_v2', body, zone)
        zone.validate()

        LOG.info('Triggered pool move for %(zone)s', {'zone': zone})
        zone = self.central_api.pool_move_zone(
            context, zone_id, target_pool_id
        )
        if zone.status == 'PENDING':
            response.status_int = 202
        else:
            response.status_int = 500

        return ''



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging
import pecan.deploy

import designate.conf


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


def factory(global_config, **local_conf):
    if not CONF['service:api'].enable_api_admin:
        def disabled_app(environ, start_response):
            status = '404 Not Found'
            start_response(status, [])
            return []

        return disabled_app

    conf = {
        'app': {
            'root': 'designate.api.admin.controllers.root.RootController',
            'modules': ['designate.api.admin'],
            'errors': {
                404: '/errors/not_found',
                405: '/errors/method_not_allowed',
                '__force_dict__': True
            }
        }
    }

    app = pecan.deploy.deploy(conf)

    return app



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import sys
from unittest import mock

from oslo_config import fixture as cfg_fixture
import oslotest.base

from designate.cmd import manage
import designate.conf
from designate.manage import base


CONF = designate.conf.CONF


class ManageTestCase(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch('oslo_config.cfg.ConfigOpts.__call__')
    @mock.patch('oslo_config.cfg.ConfigOpts.register_cli_opt')
    @mock.patch('oslo_log.log.setup')
    def test_main(self, mock_log_setup, mock_register_cli_opt,
                  mock_config_opts_call):
        script_name = 'designate-manage'
        sys.argv = [script_name, 'database', 'foo']
        action_fn = mock.MagicMock()
        CONF.category = mock.Mock(action_fn=action_fn)

        self.assertIsNone(manage.main())

        mock_register_cli_opt.assert_called()
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_config_opts_call.assert_called()
        self.assertTrue(action_fn.called)

    @mock.patch('oslo_config.cfg.ConfigOpts.__call__')
    @mock.patch('oslo_config.cfg.ConfigOpts.register_cli_opt')
    @mock.patch('oslo_log.log.setup')
    def test_main_error(self, mock_log_setup, mock_register_cli_opt,
                        mock_config_opts_call):
        script_name = 'designate-manage'
        sys.argv = [script_name, 'database', 'foo']
        action_fn = mock.MagicMock()
        CONF.category = mock.Mock(action_fn=action_fn)
        action_fn.side_effect = Exception()

        self.assertEqual(255, manage.main())

        mock_register_cli_opt.assert_called()
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_config_opts_call.assert_called()
        self.assertTrue(action_fn.called)

    def test_get_arg_string(self):
        self.assertEqual('bar', manage.get_arg_string('bar'))
        self.assertEqual('bar', manage.get_arg_string('-bar'))
        self.assertEqual('bar', manage.get_arg_string('--bar'))

    def test_methods_of(self):
        class foo:
            foo = 'bar'

            def public(self):
                pass

            def _private(self):
                pass

        methods = manage.methods_of(foo())

        method_names = [method_name for method_name, method in methods]
        self.assertIn('public', method_names)
        self.assertNotIn('_private', method_names)
        self.assertNotIn('foo', method_names)

    def test_get_available_commands(self):
        available_commands = manage.get_available_commands()

        self.assertIn('database', available_commands)
        self.assertIn('pool', available_commands)
        self.assertIn('tlds', available_commands)

    def test_fetch_func_args(self):
        expected = {
            'pool_id': mock.sentinel.pool_id,
            'current': mock.sentinel.current
        }

        CONF.category = mock.Mock(**expected)

        @base.args('--pool_id')
        @base.args('current')
        def foo():
            pass

        self.assertIn(mock.sentinel.pool_id, manage.fetch_func_args(foo))
        self.assertIn(mock.sentinel.current, manage.fetch_func_args(foo))

    def test_add_command_parsers(self):
        mock_parser = mock.Mock()
        mock_subparsers = mock.Mock()
        mock_subparsers.add_parser.return_value = mock_parser

        manage.add_command_parsers(mock_subparsers)

        mock_subparsers.add_parser.assert_any_call('database')
        mock_subparsers.add_parser.assert_any_call('pool')
        mock_subparsers.add_parser.assert_any_call('tlds')



# Copyright 2012 Bouvet ASA
#
# Author: Endre Karlson <endre.karlson@bouvet.no>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

import designate.conf
from designate.notification_handler import base


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class NeutronFloatingHandler(base.BaseAddressHandler):
    """Handler for Neutron's notifications"""
    __plugin_name__ = 'neutron_floatingip'

    def get_exchange_topics(self):
        exchange = CONF[self.name].control_exchange

        topics = [topic for topic in CONF[self.name].notification_topics]

        return (exchange, topics)

    def get_event_types(self):
        return [
            'floatingip.update.end',
            'floatingip.delete.start'
        ]

    def process_notification(self, context, event_type, payload):
        LOG.debug('%s received notification - %s',
                  self.get_canonical_name(), event_type)

        zone_id = CONF[self.name].zone_id

        if not zone_id:
            LOG.error('NeutronFloatingHandler: zone_id is None, '
                      'ignore the event.')
            return

        if event_type.startswith('floatingip.delete'):
            self._delete(zone_id=zone_id,
                         resource_id=payload['floatingip_id'],
                         resource_type='floatingip')
        elif event_type.startswith('floatingip.update'):
            if payload['floatingip']['fixed_ip_address']:
                address = {
                    'version': 4,
                    'address': payload['floatingip']['floating_ip_address']
                }
                payload['floatingip']['project'] = context.get(
                             "project_name", None)
                self._create(addresses=[address],
                             extra=payload['floatingip'],
                             zone_id=zone_id,
                             resource_id=payload['floatingip']['id'],
                             resource_type='floatingip')
            else:
                self._delete(zone_id=zone_id,
                             resource_id=payload['floatingip']['id'],
                             resource_type='floatingip')



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate import objects
from designate.objects.adapters.api_v2 import base


class ZoneTransferAcceptAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.ZoneTransferAccept
    MODIFICATIONS = {
        'fields': {
            "id": {},
            "zone_transfer_request_id": {
                'immutable': True
            },
            "project_id": {
                'rename': 'tenant_id',
            },
            "key": {
                'read_only': False
            },
            "status": {},
            "zone_id": {},
            "created_at": {},
            "updated_at": {},
        },
        'options': {
            'links': True,
            'resource_name': 'transfer_accept',
            'collection_name': 'transfer_accepts',
        }
    }

    @classmethod
    def render_object(cls, obj, *args, **kwargs):
        new_obj = super().render_object(
            obj, *args, **kwargs
        )
        new_obj['links']['zone'] = (
                '{}/v2/{}/{}'.format(
                    cls._get_base_url(kwargs['request']),
                    'zones', new_obj['zone_id']
                )
        )
        return new_obj


class ZoneTransferAcceptListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.ZoneTransferAcceptList
    MODIFICATIONS = {
        'options': {
            'links': True,
            'resource_name': 'transfer_accept',
            'collection_name': 'transfer_accepts',
        }
    }



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import uuidutils
import stevedore.exception
import yaml

from designate.backend import base as backend_base
from designate.central import rpcapi as central_rpcapi
import designate.conf
from designate import exceptions
from designate.manage import base
from designate import objects
from designate.objects.adapters import DesignateAdapter
from designate import policy
from designate import rpc


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class PoolCommands(base.Commands):
    def __init__(self):
        super().__init__()
        self.central_api = None
        self.dry_run = False
        self.skip_verify_drivers = False

    def _setup(self, dry_run=False, skip_verify_drivers=False):
        self.dry_run = dry_run
        self.skip_verify_drivers = skip_verify_drivers
        rpc.init(CONF)
        self.central_api = central_rpcapi.CentralAPI()

    @base.args('--file', help='The path to the file the yaml output should be '
                              'written to',
               default='/etc/designate/pools.yaml')
    def generate_file(self, file):
        self._setup()

        try:
            pools = self.central_api.find_pools(self.context)
            data = DesignateAdapter.render('YAML', pools)
            self._write_config_to_file(file, data)

        except messaging.exceptions.MessagingTimeout:
            LOG.critical(
                'No response received from designate-central. '
                'Check it is running, and retry'
            )
            raise SystemExit(1)

    @base.args('--pool_id', help='ID of the pool to be examined',
               default=CONF['service:central'].default_pool_id)
    @base.args('--all_pools', help='show the config of all the pools',
               default=False, required=False, action='store_true')
    def show_config(self, pool_id, all_pools):
        self._setup()

        self.output_message.append('Pool Configuration:')
        self.output_message.append('-------------------')

        try:
            pools = objects.PoolList()
            if all_pools:
                pools.extend(self.central_api.find_pools(self.context))
            else:
                if not uuidutils.is_uuid_like(pool_id):
                    self.output_message.append(
                        'Not a valid uuid: %s' % pool_id)
                    raise SystemExit(1)
                pools.append(
                    self.central_api.find_pool(self.context, {'id': pool_id}))

            self.output_message.append(
                yaml.dump(
                    DesignateAdapter.render('YAML', pools),
                    default_flow_style=False
                )
            )

        except exceptions.PoolNotFound:
            self.output_message.append('Pool not found')
            raise SystemExit(1)
        except messaging.exceptions.MessagingTimeout:
            LOG.critical(
                'No response received from designate-central. '
                'Check it is running, and retry'
            )
            raise SystemExit(1)
        finally:
            self._print_result()

    @base.args('--file', help='The path to the yaml file describing the pools',
               default='/etc/designate/pools.yaml')
    @base.args(
        '--delete',
        help='Any Pools not listed in the config file will be deleted. '
             ' WARNING: This will delete any zones left in this pool',
        action='store_true',
        default=False)
    @base.args(
        '--dry-run',
        help='This will simulate what will happen when you run this command',
        action='store_true',
        default=False)
    @base.args(
        '--skip-verify-drivers',
        help='Don\'t verify the designate backend drivers',
        action='store_true',
        default=False)
    def update(self, file, delete, dry_run=False, skip_verify_drivers=False):
        self._setup(dry_run, skip_verify_drivers)

        try:
            self.output_message.append('Updating Pools Configuration')
            self.output_message.append('****************************')

            pools_data = self._load_config(file)

            if dry_run:
                self.output_message.append('The following changes will occur:')
                self.output_message.append('*********************************')

            for pool_data in pools_data:
                try:
                    self._create_or_update_pool(pool_data)
                except exceptions.DuplicatePool:
                    raise exceptions.DuplicatePool(
                        f'Pool {pool_data["name"]} already exist with id '
                        f'{pool_data["id"]}. You cannot change id to an '
                        'existing pool.')

            if delete:
                pools = self.central_api.find_pools(self.context)
                pools_in_db = {pool.name for pool in pools}
                pools_in_yaml = {pool_data['name'] for pool_data in pools_data}
                pools_to_delete = pools_in_db - pools_in_yaml
                for pool_name in pools_to_delete:
                    self._delete_pool(pool_name)

        except exceptions.InvalidObject as e:
            self.output_message.append(str(e))
            raise SystemExit(1)
        except messaging.exceptions.MessagingTimeout:
            LOG.critical(
                'No response received from designate-central. '
                'Check it is running, and retry'
            )
            raise SystemExit(1)
        finally:
            self._print_result()

    def _create_or_update_pool(self, pool_data):
        try:
            pool = self._get_pool(pool_data)
            self._update_pool(pool_data, pool)

        except exceptions.PoolNotFound:
            self._create_pool(pool_data)

    def _get_pool(self, pool_data):
        if 'id' in pool_data:
            pool_id = pool_data['id']
            if not uuidutils.is_uuid_like(pool_id):
                self.output_message.append('Not a valid uuid: %s' % pool_id)
                raise SystemExit(1)

            pool = self.central_api.get_pool(
                self.context, pool_id
            )
        else:
            pool = self.central_api.find_pool(
                self.context, {'name': pool_data['name']}
            )

        return pool

    def _create_pool(self, pool_data):
        pool = DesignateAdapter.parse('YAML', pool_data, objects.Pool())
        self._validate_pool(pool)

        if self.dry_run:
            self.output_message.append('Create Pool: %s' % pool)
        else:
            LOG.info('Creating new pool: %s', pool)
            self.central_api.create_pool(self.context, pool)

        return pool

    def _update_pool(self, pool_data, pool):
        pool = DesignateAdapter.parse('YAML', pool_data, pool)
        self._validate_pool(pool)

        if self.dry_run:
            self.output_message.append('Update Pool: %s' % pool)
        else:
            pool = self.central_api.update_pool(self.context, pool)
            self._update_zones(pool)

    def _delete_pool(self, pool_name):
        pool = self.central_api.find_pool(
            self.context, criterion={'name': pool_name}
        )

        if self.dry_run:
            self.output_message.append('Delete Pool: %s' % pool_name)
        else:
            LOG.info('Deleting %s', pool_name)
            self.central_api.delete_pool(self.context, pool.id)

    def _update_zones(self, pool):
        LOG.info('Updating zone masters for pool: %s', pool.id)

        policy.init()
        self.context.all_tenants = True
        zones = self.central_api.find_zones(
            self.context, criterion={'pool_id': pool.id}
        )

        for zone in zones:
            zone.masters = objects.ZoneMasterList().from_list(
                self._get_masters_from_pool(pool)
            )
            self.central_api.update_zone(self.context, zone)

    def _validate_pool(self, pool):
        for ns_record in pool.ns_records:
            ns_record.validate()

        if not self.skip_verify_drivers:
            for target in pool.targets:
                try:
                    backend_base.Backend.get_driver(target.type)
                except stevedore.exception.NoMatches:
                    self.output_message.append(
                        'Unable to find designate backend driver type: '
                        '%s' % target.type
                    )
                    if not self.dry_run:
                        raise SystemExit(1)

    @staticmethod
    def _get_masters_from_pool(pool):
        masters = []
        for target in pool.targets:
            for master in target.get('masters', []):
                master = {'host': master['host'], 'port': master['port']}
                found = False
                for existing_master in masters:
                    if master == existing_master:
                        found = True
                if not found:
                    masters.append(master)
        return masters

    @staticmethod
    def _load_config(filename):
        with open(filename) as stream:
            return yaml.safe_load(stream)

    @staticmethod
    def _write_config_to_file(filename, data):
        with open(filename, 'w') as stream:
            yaml.dump(data, stream, default_flow_style=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from oslo_config import fixture as cfg_fixture
import oslotest.base

from designate.common import profiler
import designate.conf

CONF = designate.conf.CONF


class WsgiMiddlewareTest(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()

    @mock.patch.object(profiler, 'profiler_web')
    def test_factory_with_profiler_web(self, mock_profiler_web):
        mock_conf = mock.Mock()

        profiler.WsgiMiddleware.factory(mock_conf)

        mock_profiler_web.WsgiMiddleware.factory.assert_called_with(mock_conf)

    @mock.patch.object(profiler, 'profiler_web', None)
    def test_factory_without_profiler_web(self):
        mock_conf = mock.Mock()
        self.assertIsInstance(
            profiler.WsgiMiddleware.factory(mock_conf)('app'),
            profiler.WsgiMiddleware
        )

    def test_factory_call(self):
        mock_request = mock.Mock()

        profiler.WsgiMiddleware('app')(mock_request)

        mock_request.get_response.assert_called_with('app')


class ProfilerTest(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch.object(profiler, 'profiler', mock.Mock())
    @mock.patch.object(profiler, 'profiler_opts', mock.Mock())
    @mock.patch.object(profiler, 'profiler_initializer')
    def test_setup_profiler(self, mock_profiler_initializer):
        CONF.set_override('enabled', True, group='profiler')

        profiler.setup_profiler('binary', 'host')

        mock_profiler_initializer.init_from_conf.assert_called_with(
            conf=mock.ANY,
            context=mock.ANY,
            project='designate',
            service='binary',
            host='host'
        )

    @mock.patch.object(profiler, 'profiler', mock.Mock())
    @mock.patch.object(profiler, 'profiler_opts', mock.Mock())
    @mock.patch.object(profiler, 'profiler_initializer')
    def test_setup_profiler_not_enabled(self, mock_profiler_initializer):
        CONF.set_override('enabled', False, group='profiler')

        profiler.setup_profiler('binary', 'host')

        mock_profiler_initializer.init_from_conf.assert_not_called()

    @mock.patch.object(profiler, 'profiler', mock.Mock())
    @mock.patch.object(profiler, 'profiler_opts', None)
    @mock.patch.object(profiler, 'profiler_initializer')
    def test_setup_profiler_missing_dep(self, mock_profiler_initializer):
        CONF.set_override('enabled', True, group='profiler')

        profiler.setup_profiler('binary', 'host')

        mock_profiler_initializer.init_from_conf.assert_not_called()

    @mock.patch.object(profiler, 'profiler')
    def test_trace_cls(self, mock_profiler):
        CONF.set_override('enabled', True, group='profiler')

        @profiler.trace_cls('name', foo='bar')
        class Test:
            pass

        Test()

        mock_profiler.trace_cls.assert_called_with(
            'name', {'foo': 'bar'}
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import oslotest.base
from webob import exc

from designate.api.v2.controllers import rest
from designate import exceptions


class TestRestController(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()

        self.controller = rest.RestController()

    def test_handle_post(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = mock.Mock()

        self.assertEqual(
            (mock.ANY, []), self.controller._handle_post(mock.Mock(), None)
        )

    def test_handle_patch(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = mock.Mock()

        self.assertEqual(
            (mock.ANY, []), self.controller._handle_patch(mock.Mock(), None)
        )

    def test_handle_put(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = mock.Mock()

        self.assertEqual(
            (mock.ANY, []), self.controller._handle_put(mock.Mock(), None)
        )

    def test_handle_delete(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = mock.Mock()

        self.assertEqual(
            (mock.ANY, []), self.controller._handle_delete(mock.Mock(), None)
        )

    def test_handle_post_method_not_allowed(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_post, mock.Mock(), None
        )

    def test_handle_patch_method_not_allowed(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_patch, mock.Mock(), None
        )

    def test_handle_put_method_not_allowed(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_put, mock.Mock(), None
        )

    def test_handle_delete_method_not_allowed(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_delete, mock.Mock(), None
        )

    def test_handle_post_controller_not_found(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_post, mock.Mock(), ['fake']
        )

    def test_handle_patch_controller_not_found(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_patch, mock.Mock(), ['fake']
        )

    def test_handle_put_controller_not_found(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_put, mock.Mock(), ['fake']
        )

    def test_handle_delete_controller_not_found(self):
        self.controller._find_controller = mock.Mock()
        self.controller._find_controller.return_value = None

        self.assertRaisesRegex(
            exc.HTTPMethodNotAllowed,
            'The server could not comply with the request since it is either '
            'malformed or otherwise incorrect.',
            self.controller._handle_delete, mock.Mock(), ['fake']
        )


class TestFilterParams(oslotest.base.BaseTestCase):

    def test_invalid_filter_parameters(self):
        exc = self.assertRaises(
            exceptions.BadRequest,
            rest.RestController._apply_filter_params,
            {'alpha': 'foo', 'beta': 'bar'},
            ['alpha'],
            {},
        )
        self.assertIn('Invalid filters', str(exc))

    def test_duplicate_filter_parameters(self):
        exc = self.assertRaises(
            exceptions.BadRequest,
            rest.RestController._apply_filter_params,
            {'alpha': ['foo', 'baz'], 'beta': 'bar'},
            ['alpha', 'beta'],
            {},
        )
        self.assertIn('Duplicate filters', str(exc))

    def test_valid_filter_parameters(self):
        rest.RestController._apply_filter_params(
            {'alpha': 'foo', 'beta': 'bar'},
            ['alpha', 'beta'],
            {},
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

import oslotest.base

from designate.central import service
from designate.common import profiler
import designate.conf
from designate import exceptions
from designate.objects import record
from designate.objects import zone
from designate import policy
from designate import rpc


CONF = designate.conf.CONF


class CentralTestCase(oslotest.base.BaseTestCase):
    @mock.patch.object(policy, 'init')
    @mock.patch.object(rpc, 'init')
    @mock.patch.object(rpc, 'initialized')
    @mock.patch.object(profiler, 'setup_profiler')
    def setUp(self, mock_setup_profiler, mock_rpc_initialized,
              mock_rpc_init, mock_policy_init):
        super().setUp()

        mock_rpc_initialized.return_value = False

        self.storage = mock.Mock()

        self.service = service.Service()
        self.service.coordination = mock.Mock()
        self.service._storage = self.storage
        self.context = mock.Mock()

        mock_setup_profiler.assert_called()
        mock_rpc_init.assert_called()
        mock_policy_init.assert_called()

    @mock.patch.object(rpc, 'get_server')
    @mock.patch.object(rpc, 'get_notifier')
    def test_service_start(self, mock_get_notifier, mock_get_server):
        self.service.start()

        mock_get_server.assert_called()
        mock_get_notifier.assert_called_with('central')
        self.service.coordination.start.assert_called()

    @mock.patch.object(rpc, 'get_server')
    @mock.patch.object(rpc, 'get_notifier')
    def test_service_start_with_managed_tenant_id_set(self, mock_get_notifier,
                                                      mock_get_server):
        CONF.set_override(
            'managed_resource_tenant_id',
            '24d1c4be-eb2d-44bb-bc49-fc5eced95e3d',
            'service:central'
        )

        self.service.start()

        mock_get_server.assert_called()
        mock_get_notifier.assert_called_with('central')
        self.service.coordination.start.assert_called()

    def test_is_valid_project_id(self):
        self.assertIsNone(self.service._is_valid_project_id('1'))

    def test_is_valid_project_id_missing_project_id(self):
        self.assertRaisesRegex(
            exceptions.MissingProjectID,
            'A project ID must be specified when not using a project '
            'scoped token.',
            self.service._is_valid_project_id, None
        )

    def test_is_valid_zone_name_invalid_object(self):
        self.assertRaisesRegex(
            exceptions.InvalidObject,
            '',
            self.service._is_valid_zone_name, self.context, None
        )

    def test_list_to_dict(self):
        result = self.service._list_to_dict([{'region': 'foo', 'id': '1'}])
        self.assertEqual(('1',), result.popitem()[0])

    def test_check_zone_share_permission_raise_when_not_shared(self):
        new_zone = zone.Zone()
        new_zone.tenant_id = '2'
        self.context.project_id = '1'
        self.context.all_tenants = False
        self.storage.is_zone_shared_with_project.return_value = False

        self.assertRaisesRegex(
            exceptions.ZoneNotFound, 'Could not find Zone',
            self.service._check_zone_share_permission, self.context, new_zone
        )


class ZoneAndRecordStatusTestCase(oslotest.base.BaseTestCase):
    @mock.patch.object(policy, 'init')
    @mock.patch.object(rpc, 'init')
    @mock.patch.object(rpc, 'initialized')
    @mock.patch.object(profiler, 'setup_profiler')
    def setUp(self, mock_setup_profiler, mock_rpc_initialized,
              mock_rpc_init, mock_policy_init):
        super().setUp()

        mock_rpc_initialized.return_value = False

        self.service = service.Service()

        mock_setup_profiler.assert_called()
        mock_rpc_init.assert_called()
        mock_policy_init.assert_called()

    def test_success_pending(self):
        new_zone = zone.Zone(action='CREATE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'SUCCESS', 101
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('NONE', result.action)

        new_record = record.Record(action='CREATE', status='PENDING',
                                   serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'SUCCESS', 101
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('NONE', result.action)

    def test_success_error(self):
        new_zone = zone.Zone(action='CREATE', status='ERROR', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'SUCCESS', 101
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('NONE', result.action)

        new_record = record.Record(action='CREATE', status='ERROR', serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'SUCCESS', 101
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('NONE', result.action)

    def test_success_delete(self):
        new_zone = zone.Zone(action='DELETE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'SUCCESS', 101
        )

        self.assertEqual('DELETED', result.status)
        self.assertEqual('NONE', result.action)

        new_record = record.Record(action='DELETE', status='PENDING',
                                   serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'SUCCESS', 101
        )

        self.assertEqual('DELETED', result.status)
        self.assertEqual('NONE', result.action)

    def test_success_do_nothing(self):
        new_zone = zone.Zone(action='CREATE', status='PENDING', serial=2)
        result = self.service._update_zone_or_record_status(
            new_zone, 'SUCCESS', 1
        )

        self.assertEqual('PENDING', result.status)
        self.assertEqual('CREATE', result.action)

        new_record = record.Record(action='NONE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'SUCCESS', 101
        )

        self.assertEqual('PENDING', result.status)
        self.assertEqual('NONE', result.action)

    def test_error_pending(self):
        new_zone = zone.Zone(action='CREATE', status='PENDING', serial=1)
        result = self.service._update_zone_or_record_status(
            new_zone, 'ERROR', 2
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

        new_record = record.Record(action='DELETE', status='PENDING', serial=1)
        result = self.service._update_zone_or_record_status(
            new_record, 'ERROR', 2
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('DELETE', result.action)

    def test_error_pending_do_nothing(self):
        new_zone = zone.Zone(action='CREATE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'ERROR', 0
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

        new_record = record.Record(action='CREATE', status='PENDING',
                                   serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'ERROR', 0
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

    def test_error_create_do_nothing(self):
        new_zone = zone.Zone(action='CREATE', status='ACTIVE', serial=1)
        result = self.service._update_zone_or_record_status(
            new_zone, 'ERROR', 2
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('CREATE', result.action)

        new_record = record.Record(action='CREATE', status='ACTIVE', serial=1)
        result = self.service._update_zone_or_record_status(
            new_record, 'ERROR', 2
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('CREATE', result.action)

    def test_no_zone_create(self):
        new_zone = zone.Zone(action='CREATE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'NO_ZONE', 1
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

        new_record = record.Record(action='CREATE', status='PENDING',
                                   serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'NO_ZONE', 1
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

    def test_no_zone_update(self):
        new_zone = zone.Zone(action='UPDATE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'NO_ZONE', 1
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

        new_record = record.Record(action='UPDATE', status='PENDING',
                                   serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'NO_ZONE', 1
        )

        self.assertEqual('ERROR', result.status)
        self.assertEqual('CREATE', result.action)

    def test_no_zone_delete(self):
        new_zone = zone.Zone(action='DELETE', status='PENDING', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'NO_ZONE', 1
        )

        self.assertEqual('DELETED', result.status)
        self.assertEqual('NONE', result.action)

        new_record = record.Record(action='DELETE', status='PENDING',
                                   serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'NO_ZONE', 1
        )

        self.assertEqual('DELETED', result.status)
        self.assertEqual('NONE', result.action)

    def test_no_zone_do_nothing(self):
        new_zone = zone.Zone(action='NONE', status='SUCCESS', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, 'NO_ZONE', 1
        )

        self.assertEqual('SUCCESS', result.status)
        self.assertEqual('NONE', result.action)

        new_record = record.Record(action='NONE', status='ACTIVE', serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, 'NO_ZONE', 1
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('NONE', result.action)

    def test_do_nothing(self):
        new_zone = zone.Zone(action='NONE', status='SUCCESS', serial=100)
        result = self.service._update_zone_or_record_status(
            new_zone, None, None
        )

        self.assertEqual('SUCCESS', result.status)
        self.assertEqual('NONE', result.action)

        new_record = record.Record(action='NONE', status='ACTIVE', serial=100)
        result = self.service._update_zone_or_record_status(
            new_record, None, None
        )

        self.assertEqual('ACTIVE', result.status)
        self.assertEqual('NONE', result.action)



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate.objects import base
from designate.objects import fields
from designate.objects.record import Record
from designate.objects.record import RecordList


@base.DesignateRegistry.register
class SRV(Record):
    """
    SRV Resource Record Type
    Defined in: RFC2782
    """
    fields = {
        'priority': fields.IntegerFields(minimum=0, maximum=65535),
        'weight': fields.IntegerFields(minimum=0, maximum=65535),
        'port': fields.IntegerFields(minimum=0, maximum=65535),
        'target': fields.DomainField(maxLength=255),
    }

    @classmethod
    def get_recordset_schema_changes(cls):
        return {
            'name': fields.SRVField(maxLength=255, nullable=True)
        }

    def from_string(self, value):
        priority, weight, port, target = value.split(' ')
        self.priority = int(priority)
        self.weight = int(weight)
        self.port = int(port)
        self.target = target

    # The record type is defined in the RFC. This will be used when the record
    # is sent by mini-dns.
    RECORD_TYPE = 33


@base.DesignateRegistry.register
class SRVList(RecordList):

    LIST_ITEM_TYPE = SRV
    fields = {
        'objects': fields.ListOfObjectsField('SRV'),
    }



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy

from designate.common import policies
import designate.conf
from designate import exceptions


CONF = designate.conf.CONF

LOG = logging.getLogger(__name__)


_ENFORCER = None


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
    _ENFORCER = None


def set_rules(data, default_rule=None, overwrite=True):
    default_rule = default_rule or CONF.policy_default_rule
    if not _ENFORCER:
        LOG.debug("Enforcer not present, recreating at rules stage.")
        init()

    if default_rule:
        _ENFORCER.default_rule = default_rule

    msg = "Loading rules %s, default: %s, overwrite: %s"
    LOG.debug(msg, data, default_rule, overwrite)

    if isinstance(data, dict):
        rules = policy.Rules.from_dict(data, default_rule)
    else:
        rules = policy.Rules.load_json(data, default_rule)

    _ENFORCER.set_rules(rules, overwrite=overwrite)


def init(default_rule=None, policy_file=None):
    global _ENFORCER
    if not _ENFORCER:
        LOG.debug("Enforcer is not present, recreating.")
        _ENFORCER = policy.Enforcer(CONF, policy_file=policy_file)
        _ENFORCER.register_defaults(policies.list_rules())


def check(rule, ctxt, target=None, do_raise=True, exc=exceptions.Forbidden):
    creds = ctxt.to_policy_values()
    target = target or {}
    try:
        result = _ENFORCER.enforce(rule, target, creds, do_raise, exc)
    except policy.InvalidScope:
        result = False
        if do_raise:
            raise exceptions.InvalidTokenScope
    except Exception:
        result = False
        raise
    else:
        return result
    finally:
        extra = {'policy': {'rule': rule, 'target': target}}

        if result:
            LOG.trace("Policy check succeeded for rule '%(rule)s' "
                      "on target %(target)s",
                      {'rule': rule, 'target': repr(target)}, extra=extra)
        else:
            LOG.info("Policy check failed for rule '%(rule)s' "
                     "on target %(target)s",
                     {'rule': rule, 'target': repr(target)}, extra=extra)


def get_enforcer():
    # This method is used by oslopolicy CLI scripts in order to generate policy
    # files from overrides on disk and defaults in code.
    cfg.CONF([], project='designate')
    init()
    return _ENFORCER



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from designate.notification_handler import base


LOG = log.getLogger(__name__)


def get_notification_handlers(enabled_extensions):
    LOG.debug("Loading notification handlers")

    extensions = base.NotificationHandler.get_extensions(enabled_extensions)

    return [e() for e in extensions]



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import dns
import dns.query
import dns.tsigkeyring

import designate.conf
from designate import dnsutils
from designate import exceptions
from designate import objects
from designate import storage
import designate.tests.functional


CONF = designate.conf.CONF
SAMPLES = {
    ("cname.example.com.", "CNAME"): {
        "ttl": 10800,
        "records": ["example.com."],
    },
    ("_http._tcp.example.com.", "SRV"): {
        "ttl": 10800,
        "records": [
            "10 0 80 192.0.2.4.example.com.",
            "10 5 80 192.0.2.5.example.com."
        ],
    },
    ("ipv4.example.com.", "A"): {
        "ttl": 300,
        "records": ["192.0.2.1"]
    },
    ("delegation.example.com.", "NS"): {
        "ttl": 10800,
        "records": ["ns1.example.com."]
    },
    ("ipv6.example.com.", "AAAA"): {
        "ttl": 10800,
        "records": ["2001:db8::"],
    },
    ("example.com.", "SOA"): {
        "records": [
            "ns1.example.com. nsadmin.example.com."
            " 2013091101 7200 3600 2419200 10800"
        ],
        "ttl": 600
    },
    ("example.com.", "MX"): {
        "ttl": 10800,
        "records": [
            "5 192.0.2.2.example.com.",
            '10 192.0.2.3.example.com.'
        ]
    },
    ("example.com.", "TXT"): {
        "ttl": 10800,
        "records": ['"abc" "def"']
    },
    ("example.com.", "SPF"): {
        "ttl": 10800,
        "records": ['"v=spf1 mx a"']
    },
    ("example.com.", "NS"): {
        "ttl": 10800,
        "records": [
            'ns1.example.com.',
            'ns2.example.com.'
        ]
    }
}


class TestTsigUtils(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()
        self.storage = storage.get_storage()
        self.tsig_keyring = dnsutils.TsigKeyring(self.storage)

    def test_tsig_keyring(self):
        expected_result = b'J\x89\x9e:WRy\xca\xde\xb4\xa7\xb2'

        self.create_tsigkey(fixture=0)

        query = dns.message.make_query(
            'example.com.', dns.rdatatype.SOA,
        )
        query.use_tsig(dns.tsigkeyring.from_text(
            {'test-key-one': 'SomeOldSecretKey'})
        )

        self.assertEqual(expected_result, self.tsig_keyring.get(query.keyname))
        self.assertEqual(expected_result, self.tsig_keyring[query.keyname])

    def test_tsig_keyring_not_found(self):
        query = dns.message.make_query(
            'example.com.', dns.rdatatype.SOA,
        )
        query.use_tsig(dns.tsigkeyring.from_text(
            {'test-key-one': 'SomeOldSecretKey'})
        )

        self.assertIsNone(self.tsig_keyring.get(query.keyname))


class TestUtils(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()

    def test_from_dnspython_zone(self):
        zone_file = self.get_zonefile_fixture()

        dnspython_zone = dns.zone.from_text(
            zone_file,
            relativize=False,
            check_origin=False
        )

        zone = dnsutils.from_dnspython_zone(dnspython_zone)

        self.assertIsInstance(zone, objects.zone.Zone)

    def test_from_dnspython_zone_zero_soa(self):
        CONF.set_override('min_ttl', 1234, 'service:central')

        zone_file = self.get_zonefile_fixture(variant='zerosoa')

        dnspython_zone = dns.zone.from_text(
            zone_file,
            relativize=False,
            check_origin=False
        )

        zone = dnsutils.from_dnspython_zone(dnspython_zone)

        self.assertIsInstance(zone, objects.zone.Zone)
        self.assertEqual(1234, zone.ttl)

    def test_from_dnspython_zone_no_soa(self):
        zone_file = self.get_zonefile_fixture(variant='nosoa')
        dnspython_zone = dns.zone.from_text(
            zone_file,
            relativize=False,
            check_origin=False
        )

        self.assertRaisesRegex(
            exceptions.BadRequest,
            'An SOA record is required',
            dnsutils.from_dnspython_zone, dnspython_zone,
        )

    def test_parse_zone(self):
        zone_file = self.get_zonefile_fixture()

        dnspython_zone = dns.zone.from_text(
            zone_file,
            # Don't relativize, otherwise we end up with '@' record names.
            relativize=False,
            # Dont check origin, we allow missing NS records (missing SOA
            # records are taken care of in _create_zone).
            check_origin=False
        )

        zone = dnsutils.from_dnspython_zone(dnspython_zone)

        for rrset in zone.recordsets:
            k = (rrset.name, rrset.type)
            self.assertIn(k, SAMPLES)

            sample_ttl = SAMPLES[k].get('ttl', None)
            if rrset.obj_attr_is_set('ttl') or sample_ttl is not None:
                self.assertEqual(sample_ttl, rrset.ttl)

            self.assertEqual(len(rrset.records), len(SAMPLES[k]['records']))

            for record in rrset.records:
                self.assertIn(record.data, SAMPLES[k]['records'])

        self.assertEqual(len(SAMPLES), len(zone.recordsets))
        self.assertEqual('example.com.', zone.name)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

import fixtures
from oslo_log import log as logging
from oslo_messaging.rpc import dispatcher as rpc_dispatcher

from designate.central import service
import designate.conf
from designate import exceptions
from designate import objects
from designate import policy
from designate import quota
from designate import storage
from designate.storage import sqlalchemy
from designate.tests.base_fixtures import random_seed
import designate.tests.functional
from designate.tests import unit
from designate.worker import rpcapi as worker_rpcapi


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class MockZone:
    id = 1
    name = 'example.org'
    pool_id = 1
    tenant_id = 3
    ttl = 1
    type = 'PRIMARY'
    serial = 123
    shared = False

    def obj_attr_is_set(self, n):
        if n == 'recordsets':
            return False
        raise NotImplementedError()

    def __getitem__(self, k):
        items = {
            'id': 3,
            'email': 'foo@example.org',
            'serial': 123,
            'refresh': 20,
            'retry': 33,
            'expire': 9999,
            'minimum': 2,
            'name': 'example.org.',
        }
        try:
            return items[k]
        except KeyError:
            raise NotImplementedError(k)


class MockRecordSet:
    id = 1
    name = 'example.org.'
    pool_id = 1
    tenant_id = 3
    ttl = 1
    type = 'PRIMARY'
    serial = 123
    records = []

    def obj_attr_is_set(self, n):
        if n == 'records':
            return False
        raise NotImplementedError()


class MockRecord:
    hostname = 'bar'

    def __getitem__(self, n):
        assert n == 'hostname'
        return 'bar'


class MockPool:
    ns_records = [MockRecord(), ]


# Fixtures
fx_worker = fixtures.MockPatch(
    'designate.central.service.worker_rpcapi.WorkerAPI.get_instance',
    mock.MagicMock(spec_set=[
        'create_zone',
        'update_zone',
        'delete_zone'
    ])
)


class CentralBasic(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()

        pool_list = objects.PoolList.from_list(
            [
                {'id': '794ccc2c-d751-44fe-b57f-8894c9f5c842'}
            ]
        )
        attrs = {
            'count_zones.return_value': 0,
            'find_zone.return_value': MockZone(),
            'get_zone.return_value': MockZone(),
            'get_pool.return_value': MockPool(),
            'find_pools.return_value': pool_list,
        }

        self.mock_policy = mock.NonCallableMock(spec_set=[
            'reset',
            'set_rules',
            'init',
            'check',
            'enforce_new_defaults',
        ])

        self.mock_storage = mock.Mock(spec=sqlalchemy.SQLAlchemyStorage)
        self.worker_api = mock.Mock()
        self.quota = mock.Mock()
        self.mock_storage.configure_mock(**attrs)

        mock.patch.object(
            worker_rpcapi.WorkerAPI, 'get_instance',
            return_value=self.worker_api).start()
        mock.patch.object(
            policy, 'reset', return_value=mock.Mock()).start()
        mock.patch.object(
            policy, 'set_rules', return_value=mock.Mock()).start()
        mock.patch.object(
            policy, 'init', return_value=mock.Mock()).start()
        self.mock_policy_check = mock.patch.object(
            policy, 'check', return_value=mock.Mock()).start()
        self.mock_get_quota = mock.patch.object(
            quota, 'get_quota', return_value=self.quota).start()
        self.mock_get_storage = mock.patch.object(
            storage, 'get_storage', return_value=self.mock_storage).start()

        self.service = service.Service()
        self.service.check_for_tlds = True
        self.service.notifier = mock.Mock()

        self.context = mock.NonCallableMock(spec_set=[
            'elevated',
            'sudo',
            'abandon',
            'all_tenants',
            'hard_delete',
            'project_id'
        ])


class CentralServiceTestCase(CentralBasic):
    def test_conf_fixture(self):
        self.assertIn('service:central', CONF)

    def test_init(self):
        self.assertTrue(self.service.check_for_tlds)

        # Ensure these attributes are lazy
        self.mock_get_storage.assert_not_called()
        self.mock_get_quota.assert_not_called()

    def test_storage_loads_lazily(self):
        self.assertTrue(self.service.storage)
        self.mock_get_storage.assert_called_once()

    def test_quota_loads_lazily(self):
        self.assertTrue(self.service.quota)
        self.mock_get_quota.assert_called_once()

    def test_is_valid_ttl(self):
        self.CONF.set_override('min_ttl', 10, 'service:central')
        self.service._is_valid_ttl(self.context, 20)

        # policy.check() not to raise: the user is allowed to create low TTLs
        self.service._is_valid_ttl(self.context, None)
        self.service._is_valid_ttl(self.context, 1)

        # policy.check() to raise
        self.mock_policy_check.side_effect = exceptions.Forbidden

        self.assertRaisesRegex(
            exceptions.InvalidTTL,
            'TTL is below the minimum: 10',
            self.service._is_valid_ttl, self.context, 3
        )

    def test_update_soa_secondary(self):
        ctx = mock.Mock()
        mock_zone = unit.RoObject(type='SECONDARY')

        self.service._update_soa(ctx, mock_zone)
        self.assertFalse(ctx.elevated.called)

    def test_update_soa(self):
        class MockRecord:
            data = None

        mock_soa = unit.RoObject(records=[MockRecord()])

        self.context.elevated = mock.Mock()
        self.service._update_zone_in_storage = mock.Mock()
        self.service.storage.get_pool = mock.Mock(
            return_value=MockPool())
        self.service.find_recordset = mock.Mock(return_value=mock_soa)
        self.service._build_soa_record = mock.Mock()
        self.service._update_recordset_in_storage = mock.Mock()

        self.service._update_soa(self.context, MockZone())

        self.assertTrue(
            self.service._update_recordset_in_storage.called)
        self.assertTrue(self.context.elevated.called)

    def test_count_zones(self):
        self.service.count_zones(self.context)
        self.service.storage.count_zones.assert_called_once_with(
            self.context, {}
        )

    def test_count_zones_criterion(self):
        self.service.count_zones(self.context, criterion={'a': 1})
        self.service.storage.count_zones.assert_called_once_with(
            self.context, {'a': 1}
        )

    def test_validate_new_recordset(self):
        central_service = self.central_service

        central_service._is_valid_recordset_name = mock.Mock()
        central_service._is_valid_recordset_placement = mock.Mock()
        central_service._is_valid_recordset_placement_subzone = mock.Mock()
        central_service._is_valid_ttl = mock.Mock()

        MockRecordSet.id = None

        central_service._validate_recordset(
            self.context, MockZone, MockRecordSet
        )

        self.assertTrue(central_service._is_valid_recordset_name.called)
        self.assertTrue(central_service._is_valid_recordset_placement.called)
        self.assertTrue(
            central_service._is_valid_recordset_placement_subzone.called)
        self.assertTrue(central_service._is_valid_ttl.called)

    def test_validate_existing_recordset(self):
        central_service = self.central_service

        central_service._is_valid_recordset_name = mock.Mock()
        central_service._is_valid_recordset_placement = mock.Mock()
        central_service._is_valid_recordset_placement_subzone = mock.Mock()
        central_service._is_valid_ttl = mock.Mock()

        MockRecordSet.obj_get_changes = mock.Mock(return_value={'ttl': 3600})

        central_service._validate_recordset(
            self.context, MockZone, MockRecordSet
        )

        self.assertTrue(central_service._is_valid_recordset_name.called)
        self.assertTrue(central_service._is_valid_recordset_placement.called)
        self.assertTrue(
            central_service._is_valid_recordset_placement_subzone.called)
        self.assertTrue(central_service._is_valid_ttl.called)

    def test_create_recordset_in_storage(self):
        self.service._enforce_recordset_quota = mock.Mock()
        self.service._validate_recordset = mock.Mock(spec=objects.RecordSet)

        self.service.storage.create_recordset = mock.Mock(return_value='rs')
        self.service._update_zone_in_storage = mock.Mock()

        rs, zone = self.service._create_recordset_in_storage(
            self.context, MockZone(), MockRecordSet()
        )
        self.assertEqual(rs, 'rs')
        self.assertFalse(self.service._update_zone_in_storage.called)

    def test_create_recordset_with_records_in_storage(self):
        central_service = self.central_service

        central_service._enforce_recordset_quota = mock.Mock()
        central_service._enforce_record_quota = mock.Mock()
        central_service._is_valid_recordset_name = mock.Mock()
        central_service._is_valid_recordset_placement = mock.Mock()
        central_service._is_valid_recordset_placement_subzone = mock.Mock()
        central_service._is_valid_ttl = mock.Mock()

        central_service.storage.create_recordset = mock.Mock(return_value='rs')
        central_service._update_zone_in_storage = mock.Mock(
            return_value=MockZone()
        )
        recordset = mock.Mock(spec=objects.RecordSet)
        recordset.obj_attr_is_set.return_value = True
        recordset.records = [MockRecord()]

        central_service._create_recordset_in_storage(
            self.context, MockZone(), recordset
        )

        self.assertTrue(central_service._enforce_record_quota.called)
        self.assertTrue(central_service._update_zone_in_storage.called)

    def test_create_recordset_checking_DBDeadLock(self):
        self.service._enforce_recordset_quota = mock.Mock()
        self.service._enforce_record_quota = mock.Mock()
        self.service._is_valid_recordset_name = mock.Mock()
        self.service._is_valid_recordset_placement = mock.Mock()
        self.service._is_valid_recordset_placement_subzone = mock.Mock()
        self.service._is_valid_ttl = mock.Mock()

        self.service.storage._retry_on_deadlock = mock.Mock()
        self.service.storage.create_recordset = mock.Mock(return_value='rs')
        self.service._update_zone_in_storage = mock.Mock(
            return_value=MockZone()
        )

        # NOTE(thirose): Since this is a race condition we assume that
        #  we will hit it if we try to do the operations in a loop 100 times.
        for num in range(100):
            recordset = mock.Mock(spec=objects.RecordSet)
            recordset.name = f'b{num}'
            recordset.obj_attr_is_set.return_value = True
            recordset.records = [MockRecord()]

            self.service._create_recordset_in_storage(
                self.context, MockZone(), recordset
            )

            self.service.storage._retry_on_deadlock.assert_not_called()
            self.service._update_zone_in_storage.assert_called()
            self.service.storage.create_recordset.assert_called()

    def test_create_soa(self):
        self.central_service._create_recordset_in_storage = mock.Mock(
            return_value=(None, None)
        )
        self.central_service._build_soa_record = mock.Mock(
            return_value='example.org. foo.bar 1 60 5 999 1'
        )
        zone = MockZone()
        self.central_service._create_soa(self.context, zone)

        rset = (
            self.central_service._create_recordset_in_storage.call_args[0][2]
        )

        self.assertEqual('example.org.', rset.name)
        self.assertEqual('SOA', rset.type)
        self.assertEqual('SOA', rset.type)
        self.assertEqual(1, len(rset.records.objects))
        self.assertTrue(rset.records.objects[0].managed)

    def test_create_zone_in_storage(self):
        self.service._create_soa = mock.Mock()
        self.service._create_ns = mock.Mock()
        self.service.get_zone_ns_records = mock.Mock(
            return_value=[unit.RoObject(hostname='host_foo')]
        )
        self.service._ensure_catalog_zone_serial_increment = mock.Mock()

        def create_zone(ctx, zone):
            return zone

        self.service.storage.create_zone = create_zone

        zone = self.service._create_zone_in_storage(
            self.context, MockZone()
        )
        self.assertEqual('PENDING', zone.status)
        self.assertEqual('CREATE', zone.action)
        ctx, zone, hostnames = self.service._create_ns.call_args[0]
        self.assertEqual(['host_foo'], hostnames)


class CentralZoneTestCase(CentralBasic):
    zone_id = '1c85d9b0-1e9d-4e99-aede-a06664f1af2e'
    zone_id_2 = '7c85d9b0-1e9d-4e99-aede-a06664f1af2e'
    record_id = 'b81ebcfb-6236-4424-b77f-2dd0179fa041'
    record_id_2 = 'c81ebcfb-6236-4424-b77f-2dd0179fa041'
    pool_id = '769ca3fc-5924-4a44-8c1f-7efbe52fbd59'
    recordset_id = '9c85d9b0-1e9d-4e99-aede-a06664f1af2e'
    recordset_id_2 = 'dc85d9b0-1e9d-4e99-aede-a06664f1af2e'
    recordset_id_3 = '2a94a9fe-30d1-4a15-9071-0bb21996d971'
    zone_export_id = 'e887597f-9697-47dd-a202-7a2711f8669c'
    zone_shared = False

    def setUp(self):
        super().setUp()

        def storage_find_tld(c, d):
            if d['name'] not in ('org',):
                raise exceptions.TldNotFound

        def storage_find_tlds(c):
            return objects.TldList.from_list(
                [objects.Tld.from_dict({'name': 'org'})]
            )

        self.service.storage.find_tlds = storage_find_tlds
        self.service.storage.find_tld = storage_find_tld

    def test_is_valid_zone_name_valid(self):
        self.service._is_blacklisted_zone_name = mock.Mock()
        self.service._is_valid_zone_name(self.context, 'valid.org.')

    def test_is_valid_zone_name_invalid(self):
        self.service._is_blacklisted_zone_name = mock.Mock()
        self.assertRaisesRegex(
            exceptions.InvalidZoneName,
            'More than one label is required',
            self.service._is_valid_zone_name, self.context, 'example^org.'
        )

    def test_is_valid_zone_name_invalid_2(self):
        self.service._is_blacklisted_zone_name = mock.Mock()
        self.assertRaisesRegex(
            exceptions.InvalidZoneName,
            'Invalid TLD',
            self.service._is_valid_zone_name, self.context, 'example.tld.'
        )

    def test_is_valid_zone_name_invalid_same_as_tld(self):
        self.service._is_blacklisted_zone_name = mock.Mock()
        self.assertRaisesRegex(
            exceptions.InvalidZoneName,
            'Invalid TLD',
            self.service._is_valid_zone_name, self.context, 'com.com.'
        )

    def test_is_valid_zone_name_invalid_tld(self):
        self.service._is_blacklisted_zone_name = mock.Mock()
        self.assertRaisesRegex(
            exceptions.InvalidZoneName,
            'More than one label is required',
            self.service._is_valid_zone_name, self.context, 'tld.'
        )

    def test_is_valid_zone_name_blacklisted(self):
        self.service._is_blacklisted_zone_name = mock.Mock(
            side_effect=exceptions.InvalidZoneName)
        self.assertRaisesRegex(
            exceptions.InvalidZoneName,
            'Invalid TLD',
            self.service._is_valid_zone_name, self.context, 'valid.com.'
        )

    def test_is_blacklisted_zone_name(self):
        self.service.storage.find_blacklists.return_value = [
            unit.RoObject(pattern='a'), unit.RoObject(pattern='b')
        ]
        blacklist_tests = (
            ('example.org', True),
            ('example.net', True),
            ('hi', False),
            ('', False)
        )
        for zone, expected in blacklist_tests:
            self.assertEqual(
                self.service._is_blacklisted_zone_name(self.context, zone),
                expected
            )

    def test_is_valid_recordset_name(self):
        zone = unit.RoObject(name='example.org.')
        self.service._is_valid_recordset_name(self.context, zone,
                                              'foo..example.org.')

    def test_is_valid_recordset_name_no_dot(self):
        zone = unit.RoObject(name='example.org.')
        self.assertRaisesRegex(
            ValueError,
            'Please supply a FQDN',
            self.service._is_valid_recordset_name,
            self.context, zone, 'foo.example.org',
        )

    def test_is_valid_recordset_name_too_long(self):
        zone = unit.RoObject(name='example.org.')
        CONF['service:central'].max_recordset_name_len = 255
        rs_name = 'a' * 255 + '.org.'
        self.assertRaisesRegex(
            exceptions.InvalidRecordSetName,
            'Name too long',
            self.service._is_valid_recordset_name,
            self.context, zone, rs_name,
        )

    def test_is_valid_recordset_name_wrong_zone(self):
        zone = unit.RoObject(name='example.org.')
        self.assertRaisesRegex(
            exceptions.InvalidRecordSetLocation,
            'RecordSet is not contained within it\'s parent zone',
            self.service._is_valid_recordset_name,
            self.context, zone, 'foo.example.com.',
        )

    def test_is_valid_recordset_placement_cname(self):
        zone = unit.RoObject(name='example.org.')
        self.assertRaisesRegex(
            exceptions.InvalidRecordSetLocation,
            'CNAME recordsets may not be created at the zone apex',
            self.service._is_valid_recordset_placement,
            self.context, zone, 'example.org.', 'CNAME',
        )

    def test_is_valid_recordset_placement_failing(self):
        zone = unit.RoObject(
            name='example.org.', id=CentralZoneTestCase.zone_id
        )
        self.service.storage.find_recordsets.return_value = [
            unit.RoObject(id=CentralZoneTestCase.recordset_id)
        ]
        self.assertRaisesRegex(
            exceptions.InvalidRecordSetLocation,
            'CNAME recordsets may not share a name with any other records',
            self.service._is_valid_recordset_placement,
            self.context, zone, 'example.org.', 'A',
        )

    def test_is_valid_recordset_placement_failing_2(self):
        zone = unit.RoObject(
            name='example.org.', id=CentralZoneTestCase.zone_id
        )
        self.service.storage.find_recordsets.return_value = [
            unit.RoObject(),
            unit.RoObject()
        ]
        self.assertRaisesRegex(
            exceptions.InvalidRecordSetLocation,
            'CNAME recordsets may not share a name with any other records',
            self.service._is_valid_recordset_placement,
            self.context, zone, 'example.org.', 'A',
        )

    def test_is_valid_recordset_placement(self):
        zone = unit.RoObject(
            name='example.org.', id=CentralZoneTestCase.zone_id
        )
        self.service.storage.find_recordsets.return_value = []
        ret = self.service._is_valid_recordset_placement(
            self.context,
            zone,
            'example.org.',
            'A',
        )
        self.assertTrue(ret)

    def test_is_valid_recordset_placement_subzone(self):
        zone = unit.RoObject(
            name='example.org.', id=CentralZoneTestCase.zone_id
        )
        self.service._is_valid_recordset_placement_subzone(
            self.context,
            zone,
            'example.org.'
        )

    def test_is_valid_recordset_placement_subzone_2(self):
        zone = unit.RoObject(
            name='example.org.', id=CentralZoneTestCase.zone_id
        )
        self.service._is_valid_recordset_name = mock.Mock(
            side_effect=Exception)
        self.service.storage.find_zones.return_value = [
            unit.RoObject(name='foo.example.org.')
        ]
        self.service._is_valid_recordset_placement_subzone(
            self.context,
            zone,
            'bar.example.org.'
        )

    def test_is_valid_recordset_placement_subzone_failing(self):
        zone = unit.RoObject(
            name='example.org.', id=CentralZoneTestCase.zone_id
        )
        self.service._is_valid_recordset_name = mock.Mock()
        self.service.storage.find_zones.return_value = [
            unit.RoObject(name='foo.example.org.')
        ]

        self.assertRaisesRegex(
            exceptions.InvalidRecordSetLocation,
            'RecordSet belongs in a child zone: foo.example.org.',
            self.service._is_valid_recordset_placement_subzone,
            self.context, zone, 'bar.example.org.'
        )

    def test_is_valid_recordset_records(self):
        recordset = unit.RoObject(
            records=[
                'ww1.example.com.',
                'ww2.example.com.'
            ],
            type='CNAME'
        )
        self.assertRaisesRegex(
            exceptions.BadRequest,
            'CNAME recordsets may not have more than 1 record',
            self.service._is_valid_recordset_records, recordset
        )

    def test_is_superzone(self):
        central_service = self.central_service

        central_service.storage.find_zones = mock.Mock()
        central_service._is_superzone(self.context, 'example.org.', '1')
        _, crit = self.service.storage.find_zones.call_args[0]
        self.assertEqual({'name': '%.example.org.', 'pool_id': '1'}, crit)

    def test_create_ns(self):
        self.service._create_recordset_in_storage = mock.Mock(
            return_value=(0, 0))
        self.service._create_ns(
            self.context,
            unit.RoObject(type='PRIMARY', name='example.org.'),
            [unit.RoObject(), unit.RoObject(), unit.RoObject()]
        )
        ctx, zone, rset = (
            self.service._create_recordset_in_storage.call_args[0])

        self.assertEqual('example.org.', rset.name)
        self.assertEqual('NS', rset.type)
        self.assertEqual(3, len(rset.records))
        self.assertTrue(rset.records[0].managed)

    def test_create_ns_skip(self):
        self.service._create_recordset_in_storage = mock.Mock()
        self.service._create_ns(
            self.context,
            unit.RoObject(type='SECONDARY', name='example.org.'),
            [],
        )
        self.assertFalse(
            self.service._create_recordset_in_storage.called)

    def test_add_ns_creation(self):
        self.service._create_ns = mock.Mock()

        self.service.find_recordset = mock.Mock(
            side_effect=exceptions.RecordSetNotFound()
        )

        self.service._add_ns(
            self.context,
            unit.RoObject(name='foo', id=CentralZoneTestCase.zone_id),
            unit.RoObject(name='bar')
        )
        ctx, zone, records = self.service._create_ns.call_args[0]
        self.assertTrue(len(records), 1)

    def test_add_ns(self):
        self.service._update_recordset_in_storage = mock.Mock()

        self.service.find_recordset = mock.Mock(
            return_value=unit.RoObject(
                records=objects.RecordList.from_list([]), managed=True
            )
        )

        self.service._add_ns(
            self.context,
            unit.RoObject(name='foo', id=CentralZoneTestCase.zone_id),
            unit.RoObject(name='bar')
        )
        ctx, zone, rset = (
            self.service._update_recordset_in_storage.call_args[0])
        self.assertEqual(len(rset.records), 1)
        self.assertTrue(rset.records[0].managed)
        self.assertEqual('bar', rset.records[0].data.name)

    def test_create_zone_no_servers(self):
        self.service._enforce_zone_quota = mock.Mock()
        self.service._is_valid_zone_name = mock.Mock()
        self.service._is_valid_ttl = mock.Mock()
        self.service._is_subzone = mock.Mock(
            return_value=False
        )
        self.service._is_superzone = mock.Mock(
            return_value=[]
        )
        self.service.storage.get_pool.return_value = unit.RoObject(
            ns_records=[]
        )

        self.useFixture(
            fixtures.MockPatchObject(
                self.service.storage,
                'find_pools',
                return_value=objects.PoolList.from_list(
                    [
                        {'id': '94ccc2c1-d751-44fe-b57f-8894c9f5c842'}
                    ]
                )
            )
        )

        z = objects.Zone(tenant_id='1',
                         name='example.com.', ttl=60,
                         pool_id=CentralZoneTestCase.pool_id)

        self.assertRaises(exceptions.NoServersConfigured,
                          self.service.create_zone,
                          self.context, z)

    def test_create_zone(self):
        self.service._enforce_zone_quota = mock.Mock()
        self.service._create_zone_in_storage = mock.Mock(
            return_value=objects.Zone(
                name='example.com.',
                type='PRIMARY',
            )
        )
        self.service._is_valid_zone_name = mock.Mock()
        self.service._is_valid_ttl = mock.Mock()
        self.service._is_subzone = mock.Mock(
            return_value=False
        )
        self.service._is_superzone = mock.Mock(
            return_value=[]
        )
        self.service.storage.get_pool.return_value = unit.RoObject(
            ns_records=[unit.RoObject()]
        )
        self.useFixture(
            fixtures.MockPatchObject(
                self.service.storage,
                'find_pools',
                return_value=objects.PoolList.from_list(
                    [
                        {'id': '94ccc2c1-d751-44fe-b57f-8894c9f5c842'}
                    ]
                )
            )
        )
        self.service.storage.get_catalog_zone = mock.Mock(
            side_effect=exceptions.ZoneNotFound)

        out = self.service.create_zone(
            self.context,
            objects.Zone(
                tenant_id='1',
                name='example.com.',
                ttl=60,
                pool_id=CentralZoneTestCase.pool_id,
                refresh=0,
                type='PRIMARY'
            )
        )
        self.assertEqual('example.com.', out.name)

    def test_get_zone(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='foo',
            tenant_id='2',
            shared=self.zone_shared,
        )
        self.service.get_zone(self.context,
                              CentralZoneTestCase.zone_id)

        self.mock_policy_check.assert_called_with(
            'get_zone', mock.ANY,
            {
                'zone_id': CentralZoneTestCase.zone_id,
                'zone_name': 'foo',
                'zone_shared': False,
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    def test_get_zone_servers(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='foo',
            tenant_id='2',
            pool_id=CentralZoneTestCase.pool_id,
        )

        self.service.get_zone_ns_records(
            self.context,
            zone_id=CentralZoneTestCase.zone_id
        )

        ctx, pool_id = self.service.storage.get_pool.call_args[0]
        self.assertEqual(CentralZoneTestCase.pool_id, pool_id)

    def test_find_zones(self):
        self.context = unit.RoObject(project_id='t', roles=[])
        self.service.storage.find_zones = mock.Mock()
        self.service.find_zones(self.context)
        self.assertTrue(self.service.storage.find_zones.called)

        self.mock_policy_check.assert_called_with(
            'find_zones', mock.ANY, {'project_id': 't', 'tenant_id': 't'}
        )

    def test_delete_zone_has_subzone(self):
        self.context.abandon = False
        self.context.hard_delete = False
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='foo',
            tenant_id='2',
            shared=self.zone_shared,
            type='PRIMARY',
        )
        self.service.storage.count_zones.return_value = 2

        self.assertRaises(exceptions.ZoneHasSubZone,
                          self.service.delete_zone,
                          self.context,
                          CentralZoneTestCase.zone_id)

        self.mock_policy_check.assert_called_with(
            'delete_zone', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id,
                'zone_name': 'foo',
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    @mock.patch.object(designate.central.service.Service,
                       '_ensure_catalog_zone_serial_increment')
    def test_delete_zone_abandon(
            self, mock_ensure_catalog_zone_serial_increment):
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='foo',
            tenant_id='2',
            id=CentralZoneTestCase.zone_id_2,
            shared=self.zone_shared,
            type='PRIMARY',
        )
        self.context.abandon = True
        self.service.storage.count_zones.return_value = 0
        self.service.delete_zone(self.context,
                                 CentralZoneTestCase.zone_id)
        self.assertTrue(self.service.storage.delete_zone.called)
        self.assertFalse(self.service.worker_api.delete_zone.called)

        self.mock_policy_check.assert_called_with(
            'abandon_zone', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id,
                'zone_name': 'foo',
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    # RoObject not compatible with _ensure_catalog_zone_serial_increment
    @mock.patch.object(designate.central.service.Service,
                       '_ensure_catalog_zone_serial_increment')
    def test_delete_zone(self, mock_ensure_catalog_zone_serial_increment):
        self.context.abandon = False
        self.context.hard_delete = False
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='foo',
            tenant_id='2',
            shared=self.zone_shared,
            type='PRIMARY',
        )
        self.service._delete_zone_in_storage = mock.Mock(
            return_value=unit.RoObject(
                name='foo'
            )
        )
        self.service.storage.count_zones.return_value = 0
        out = self.service.delete_zone(self.context,
                                       CentralZoneTestCase.zone_id)
        self.assertFalse(self.service.storage.delete_zone.called)
        self.assertTrue(self.service.worker_api.delete_zone.called)

        self.mock_policy_check.assert_called()
        ctx, deleted_dom = self.service.worker_api.delete_zone.call_args[0]

        self.assertEqual('foo', deleted_dom.name)
        self.assertEqual('foo', out.name)
        self.mock_policy_check.assert_called_with(
            'delete_zone', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id,
                'zone_name': 'foo',
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    # RoObject not compatible with _ensure_catalog_zone_serial_increment
    @mock.patch.object(designate.central.service.Service,
                       '_ensure_catalog_zone_serial_increment')
    def test_delete_zone_hard_delete(
            self, mock_ensure_catalog_zone_serial_increment):
        self.context.abandon = False
        self.context.hard_delete = True
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='foo',
            tenant_id='2',
            shared=False,
            type='PRIMARY',
        )
        self.service._delete_zone_in_storage = mock.Mock(
            return_value=unit.RoObject(
                name='foo'
            )
        )
        self.service.storage.count_zones.return_value = 0
        out = self.service.delete_zone(self.context,
                                       CentralZoneTestCase.zone_id)
        self.assertFalse(self.service.storage.delete_zone.called)
        self.assertTrue(self.service.worker_api.delete_zone.called)
        self.mock_policy_check.assert_called()
        ctx, deleted_dom = (
            self.service.worker_api.delete_zone.call_args[0])

        self.assertEqual('foo', deleted_dom.name)
        self.assertEqual('foo', out.name)
        self.mock_policy_check.assert_called_with(
            'delete_zone', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id,
                'zone_name': 'foo',
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    def test_delete_zone_in_storage(self):
        self.service._ensure_catalog_zone_serial_increment = mock.Mock()
        self.service._delete_zone_in_storage(
            self.context,
            unit.RwObject(action='', status=''),
        )
        d = self.service.storage.update_zone.call_args[0][1]
        self.assertEqual('DELETE', d.action)
        self.assertEqual('PENDING', d.status)

    def test_xfr_zone_secondary(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='example.org.',
            tenant_id='2',
            type='SECONDARY',
            masters=[unit.RoObject(host='192.0.2.1', port=53)],
            serial=1,
        )
        with fx_worker:
            self.service.worker_api.get_serial_number.return_value = (
                'SUCCESS', 2
            )
            self.service.xfr_zone(
                self.context, CentralZoneTestCase.zone_id)
            self.assertTrue(
                self.service.worker_api.perform_zone_xfr.called)

            self.mock_policy_check.assert_called()
        self.mock_policy_check.assert_called_with(
            'xfr_zone', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id,
                'zone_name': 'example.org.',
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    def test_xfr_zone_not_secondary(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            name='example.org.',
            tenant_id='2',
            type='PRIMARY',
        )

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.xfr_zone,
                                self.context,
                                CentralZoneTestCase.zone_id)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

    def test_count_report(self):
        self.service.count_zones = mock.Mock(return_value=1)
        self.service.count_records = mock.Mock(return_value=2)
        self.service.count_tenants = mock.Mock(return_value=3)
        reports = self.service.count_report(
            self.context,
            criterion=None
        )
        self.assertEqual([{'zones': 1, 'records': 2, 'tenants': 3}], reports)

    def test_count_report_zones(self):
        self.service.count_zones = mock.Mock(return_value=1)
        self.service.count_records = mock.Mock(return_value=2)
        self.service.count_tenants = mock.Mock(return_value=3)
        reports = self.service.count_report(
            self.context,
            criterion='zones'
        )
        self.assertEqual([{'zones': 1}], reports)

    def test_count_report_records(self):
        self.service.count_zones = mock.Mock(return_value=1)
        self.service.count_records = mock.Mock(return_value=2)
        self.service.count_tenants = mock.Mock(return_value=3)
        reports = self.service.count_report(
            self.context,
            criterion='records'
        )
        self.assertEqual([{'records': 2}], reports)

    def test_count_report_tenants(self):
        self.service.count_zones = mock.Mock(return_value=1)
        self.service.count_records = mock.Mock(return_value=2)
        self.service.count_tenants = mock.Mock(return_value=3)
        reports = self.service.count_report(
            self.context,
            criterion='tenants'
        )
        self.assertEqual([{'tenants': 3}], reports)

    def test_count_report_not_found(self):
        self.service.count_zones = mock.Mock(return_value=1)
        self.service.count_records = mock.Mock(return_value=2)
        self.service.count_tenants = mock.Mock(return_value=3)

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.count_report,
                                self.context,
                                criterion='bogus')

        self.assertEqual(exceptions.ReportNotFound, exc.exc_info[0])

    def test_get_recordset_not_found(self):
        zone = MockZone()
        zone.id = CentralZoneTestCase.zone_id
        self.service.storage.get_zone.return_value = zone
        self.service.storage.find_recordset.return_value = unit.RoObject(
            zone_id=CentralZoneTestCase.zone_id_2
        )

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.get_recordset,
                                self.context,
                                CentralZoneTestCase.zone_id,
                                CentralZoneTestCase.recordset_id)

        self.assertEqual(exceptions.RecordSetNotFound, exc.exc_info[0])

    def test_get_recordset(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            id=CentralZoneTestCase.zone_id_2,
            name='example.org.',
            tenant_id='2',
            shared=self.zone_shared,
        )
        recordset = objects.RecordSet(
            zone_id=CentralZoneTestCase.zone_id_2,
            zone_name='example.org.',
            id=CentralZoneTestCase.recordset_id
        )

        self.service.storage.find_recordset.return_value = recordset

        self.service.get_recordset(
            self.context,
            CentralZoneTestCase.zone_id_2,
            CentralZoneTestCase.recordset_id,
        )

        self.mock_policy_check.assert_called_with(
            'get_recordset', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id_2,
                'zone_name': 'example.org.',
                'zone_shared': False,
                'recordset_id': CentralZoneTestCase.recordset_id,
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    def test_get_recordset_no_zone_id(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            id=CentralZoneTestCase.zone_id_2,
            name='example.org.',
            tenant_id='2',
            shared=self.zone_shared,
        )
        recordset = objects.RecordSet(
            zone_id=CentralZoneTestCase.zone_id_2,
            zone_name='example.org.',
            id=CentralZoneTestCase.recordset_id
        )

        self.service.storage.find_recordset.return_value = recordset

        # Set the zone_id value to false
        self.service.get_recordset(
            self.context,
            False,
            CentralZoneTestCase.recordset_id,
        )

        self.mock_policy_check.assert_called_with(
            'get_recordset', mock.ANY, {
                'zone_id': CentralZoneTestCase.zone_id_2,
                'zone_name': 'example.org.',
                'zone_shared': False,
                'recordset_id': CentralZoneTestCase.recordset_id,
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    def test_find_recordsets(self):
        self.context = mock.Mock()
        self.context.project_id = 't'
        self.service.find_recordsets(self.context)
        self.assertTrue(self.service.storage.find_recordsets.called)

        self.mock_policy_check.assert_called_with(
            'find_recordsets', mock.ANY, {'project_id': 't', 'tenant_id': 't'}
        )

    def test_find_recordset(self):
        self.context = mock.Mock()
        self.context.project_id = 't'
        self.service.storage.get_zone.return_value = MockZone()
        self.service.find_recordset(self.context)
        self.assertTrue(self.service.storage.find_recordset.called)
        self.mock_policy_check.assert_called_with(
            'find_recordset', mock.ANY, {'project_id': 't', 'tenant_id': 't'}
        )

    def test_update_recordset_fail_on_changes(self):
        self.service.storage.get_zone.return_value = unit.RoObject()
        recordset = mock.Mock(spec=objects.RecordSet)
        recordset.obj_get_original_value.return_value = '1'

        recordset.obj_get_changes.return_value = ['tenant_id', 'foo']
        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.update_recordset,
                                self.context,
                                recordset)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

        recordset.obj_get_changes.return_value = ['zone_id', 'foo']
        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.update_recordset,
                                self.context,
                                recordset)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

        recordset.obj_get_changes.return_value = ['type', 'foo']
        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.update_recordset,
                                self.context,
                                recordset)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

    def test_update_recordset_action_delete(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            action='DELETE', tenant_id='', type='PRIMARY'
        )
        recordset = mock.Mock(spec=objects.RecordSet)
        recordset.obj_get_changes.return_value = ['foo']

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.update_recordset,
                                self.context,
                                recordset)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

    def test_update_recordset_action_fail_on_managed(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            type='foo',
            name='example.org.',
            tenant_id='2',
            action='bogus',
            shared=self.zone_shared,
        )
        recordset = mock.Mock(spec=objects.RecordSet)
        recordset.obj_get_changes.return_value = ['foo']
        recordset.managed = True
        self.context = mock.Mock()
        self.context.edit_managed_records = False

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.update_recordset,
                                self.context,
                                recordset)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

    def test_update_recordset_worker_model(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            type='foo',
            name='example.org.',
            tenant_id='2',
            action='bogus',
            shared=self.zone_shared,
        )
        recordset = mock.Mock(spec=objects.RecordSet)
        recordset.obj_get_changes.return_value = ['foo']
        recordset.obj_get_original_value.return_value = (
            '9c85d9b0-1e9d-4e99-aede-a06664f1af2e'
        )
        recordset.managed = False
        self.service._update_recordset_in_storage = mock.Mock(
            return_value=('x', 'y')
        )

        with fx_worker:
            self.service.update_recordset(self.context, recordset)
        self.assertTrue(
            self.service._update_recordset_in_storage.called)

        self.mock_policy_check.assert_called_with(
            'update_recordset', mock.ANY, {
                'recordset_id': '9c85d9b0-1e9d-4e99-aede-a06664f1af2e',
                'recordset_project_id': '9c85d9b0-1e9d-4e99-aede-a06664f1af2e',
                'zone_id': '9c85d9b0-1e9d-4e99-aede-a06664f1af2e',
                'zone_name': 'example.org.',
                'zone_shared': self.zone_shared,
                'zone_type': 'foo',
                'project_id': '2',
                'tenant_id': '2'
            }
        )

    def test_update_recordset_in_storage(self):
        recordset = mock.Mock()
        recordset.name = 'n'
        recordset.type = 't'
        recordset.id = CentralZoneTestCase.recordset_id
        recordset.obj_get_changes.return_value = {'ttl': 90}
        recordset.ttl = 90
        recordset.records = []
        self.service._is_valid_recordset_name = mock.Mock()
        self.service._is_valid_recordset_placement = mock.Mock()
        self.service._is_valid_recordset_placement_subzone = mock.Mock()
        self.service._is_valid_ttl = mock.Mock()
        self.service._update_zone_in_storage = mock.Mock()

        self.service._update_recordset_in_storage(
            self.context,
            unit.RoObject(serial=3),
            recordset,
        )

        self.assertEqual(
            'n',
            self.service._is_valid_recordset_name.call_args[0][2]
        )
        self.assertEqual(
            ('n', 't', CentralZoneTestCase.recordset_id),
            self.service._is_valid_recordset_placement.call_args[0][2:]
        )
        self.assertEqual(
            'n',
            self.service._is_valid_recordset_placement_subzone.
            call_args[0][2]
        )
        self.assertEqual(
            90,
            self.service._is_valid_ttl.call_args[0][1]
        )
        self.assertTrue(self.service.storage.update_recordset.called)
        self.assertTrue(self.service._update_zone_in_storage.called)

    def test_update_recordset_in_storage_2(self):
        recordset = mock.Mock()
        recordset.name = 'n'
        recordset.type = 't'
        recordset.id = CentralZoneTestCase.recordset_id
        recordset.ttl = None
        recordset.obj_get_changes.return_value = {'ttl': None}
        recordset.records = [unit.RwObject(
            action='a',
            status='s',
            serial=9,
        )]
        self.service._is_valid_recordset_name = mock.Mock()
        self.service._is_valid_recordset_placement = mock.Mock()
        self.service._is_valid_recordset_placement_subzone = mock.Mock()
        self.service._update_zone_in_storage = mock.Mock()
        self.service._enforce_record_quota = mock.Mock()

        self.service._update_recordset_in_storage(
            self.context,
            unit.RoObject(serial=3),
            recordset,
            increment_serial=False,
        )

        self.assertEqual(
            'n',
            self.service._is_valid_recordset_name.call_args[0][2]
        )
        self.assertEqual(
            ('n', 't', CentralZoneTestCase.recordset_id),
            self.service._is_valid_recordset_placement.call_args[0][2:]
        )
        self.assertEqual(
            'n',
            self.service._is_valid_recordset_placement_subzone.
            call_args[0][2]
        )
        self.assertFalse(self.service._update_zone_in_storage.called)
        self.assertTrue(self.service.storage.update_recordset.called)
        self.assertTrue(self.service._enforce_record_quota.called)

    def test_delete_recordset_not_found(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            action='bogus',
            id=CentralZoneTestCase.zone_id_2,
            name='example.org.',
            tenant_id='2',
            type='foo',
            shared=self.zone_shared,
        )
        self.service.storage.find_recordset.side_effect = (
            exceptions.RecordSetNotFound()
        )
        self.context = mock.Mock()
        self.context.edit_managed_records = False

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.delete_recordset,
                                self.context,
                                CentralZoneTestCase.zone_id_2,
                                CentralZoneTestCase.recordset_id)

        self.assertEqual(exceptions.RecordSetNotFound, exc.exc_info[0])

    def test_delete_recordset_action_delete(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            action='DELETE',
            id=CentralZoneTestCase.zone_id_2,
            name='example.org.',
            tenant_id='2',
            type='foo',
        )
        self.service.storage.find_recordset.return_value = unit.RoObject(
            zone_id=CentralZoneTestCase.zone_id_2,
            id=CentralZoneTestCase.recordset_id,
            managed=False,
        )
        self.context = mock.Mock()
        self.context.edit_managed_records = False

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.delete_recordset,
                                self.context,
                                CentralZoneTestCase.zone_id_2,
                                CentralZoneTestCase.recordset_id)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

    def test_delete_recordset_managed(self):
        self.service.storage.get_zone.return_value = unit.RoObject(
            action='foo',
            id=CentralZoneTestCase.zone_id_2,
            name='example.org.',
            tenant_id='2',
            type='foo',
            shared=self.zone_shared,
        )
        self.service.storage.find_recordset.return_value = unit.RoObject(
            zone_id=CentralZoneTestCase.zone_id_2,
            id=CentralZoneTestCase.recordset_id,
            managed=True,
            tenant_id='2',
        )
        self.context = mock.Mock()
        self.context.edit_managed_records = False

        exc = self.assertRaises(rpc_dispatcher.ExpectedException,
                                self.service.delete_recordset,
                                self.context,
                                CentralZoneTestCase.zone_id_2,
                                CentralZoneTestCase.recordset_id)

        self.assertEqual(exceptions.BadRequest, exc.exc_info[0])

    def test_delete_recordset_worker(self):
        mock_zone = unit.RoObject(
            action='foo',
            id=CentralZoneTestCase.zone_id_2,
            name='example.org.',
            tenant_id='2',
            type='foo',
            shared=self.zone_shared,
        )
        mock_rs = objects.RecordSet(
            zone_id=CentralZoneTestCase.zone_id_2,
            zone_name='example.org.',
            id=CentralZoneTestCase.recordset_id,
            records=objects.RecordList.from_list([]),
        )

        self.service.storage.get_zone.return_value = mock_zone
        self.service.storage.find_recordset.return_value = mock_rs
        self.context = mock.Mock()
        self.context.edit_managed_records = False
        self.service._delete_recordset_in_storage = mock.Mock(
            return_value=(mock_rs, mock_zone)
        )

        with fx_worker:
            self.service.delete_recordset(self.context,
                                          CentralZoneTestCase.zone_id_2,
                                          CentralZoneTestCase.recordset_id)

        self.assertTrue(
            self.service._delete_recordset_in_storage.called)

    def test_delete_recordset_in_storage(self):
        def mock_uds(c, zone, inc):
            return zone

        self.service._update_zone_in_storage = mock_uds
        self.service._delete_recordset_in_storage(
            self.context,
            unit.RoObject(serial=1, shared=self.zone_shared),
            unit.RoObject(id=2, records=[
                unit.RwObject(
                    action='',
                    status='',
                    serial=0,
                    increment_serial=False,
                )
            ])
        )
        self.assertTrue(self.service.storage.update_recordset.called)
        self.assertTrue(self.service.storage.delete_recordset.called)
        rs = self.service.storage.update_recordset.call_args[0][1]
        self.assertEqual(1, len(rs.records))
        self.assertEqual('DELETE', rs.records[0].action)
        self.assertEqual('PENDING', rs.records[0].status)
        self.assertTrue(rs.records[0].serial, 1)

    def test_delete_recordset_in_storage_no_increment_serial(self):
        self.service._update_zone_in_storage = mock.Mock()
        self.service._delete_recordset_in_storage(
            self.context,
            unit.RoObject(serial=1, shared=self.zone_shared),
            unit.RoObject(id=2, records=[
                unit.RwObject(
                    action='',
                    status='',
                    serial=0,
                )
            ]),
            increment_serial=False,
        )
        self.assertTrue(self.service.storage.update_recordset.called)
        self.assertTrue(self.service.storage.delete_recordset.called)
        self.assertFalse(self.service._update_zone_in_storage.called)

    def test_count_recordset(self):
        self.service.count_recordsets(self.context)
        self.mock_policy_check.assert_called_with(
            'count_recordsets', mock.ANY, {
                'project_id': None, 'tenant_id': None}
        )

    def test_count_records(self):
        self.service.count_records(self.context)
        self.mock_policy_check.assert_called_with(
            'count_records', mock.ANY, {'project_id': None, 'tenant_id': None}
        )

    def test_determine_floatingips(self):
        self.context = mock.Mock()
        self.context.project_id = 'tnt'
        self.service.find_records = mock.Mock(return_value=[
            unit.RoObject(managed_extra='')
        ])

        fips = {}
        data, invalid = self.service._determine_floatingips(
            self.context, fips)
        self.assertEqual({}, data)
        self.assertEqual([], invalid)

    def test_determine_floatingips_with_data(self):
        self.context = mock.Mock()
        self.context.project_id = 2
        self.service.find_records = mock.Mock(return_value=[
            unit.RoObject(managed_extra=1, managed_tenant_id=1),
            unit.RoObject(managed_extra=2, managed_tenant_id=2),
        ])

        fips = {
            'k': {'address': 1},
            'k2': {'address': 2},
        }
        data, invalid = self.service._determine_floatingips(
            self.context, fips)
        self.assertEqual(1, len(invalid))
        self.assertEqual(1, invalid[0].managed_tenant_id)
        self.assertEqual(data['k'], ({'address': 1}, None))

    def test_generate_soa_refresh_interval(self):
        central_service = self.central_service
        with random_seed(42):
            refresh_time = central_service._generate_soa_refresh_interval()
            self.assertEqual(3563, refresh_time)


class IsSubzoneTestCase(CentralBasic):
    def setUp(self):
        super().setUp()

        def find_zone(ctx, criterion):
            LOG.debug('Calling find_zone on %r' % criterion)
            if criterion['name'] == 'example.com.':
                LOG.debug('Returning %r' % criterion['name'])
                return criterion['name']

            LOG.debug('Not found')
            raise exceptions.ZoneNotFound

        self.service.storage.find_zone = find_zone

    def test_is_subzone_false(self):
        r = self.service._is_subzone(self.context, 'com',
                                     CentralZoneTestCase.pool_id)
        self.assertFalse(r)

    def FIXME_test_is_subzone_false2(self):
        r = self.service._is_subzone(self.context, 'com.',
                                     CentralZoneTestCase.pool_id)
        self.assertEqual('com.', r)

    def FIXME_test_is_subzone_false3(self):
        r = self.service._is_subzone(self.context, 'example.com.',
                                     CentralZoneTestCase.pool_id)
        self.assertEqual('example.com.', r)

    def test_is_subzone_false4(self):
        r = self.service._is_subzone(
            self.context, 'foo.a.b.example.com.',
            CentralZoneTestCase.pool_id)
        self.assertEqual('example.com.', r)


class CentralZoneExportTests(CentralBasic):
    def setUp(self):
        super().setUp()

        def storage_find_tld(c, d):
            if d['name'] not in ('org',):
                raise exceptions.TldNotFound

        self.service.storage.find_tld = storage_find_tld

    def test_create_zone_export(self):
        self.context = mock.Mock()
        self.context.project_id = 't'

        self.service.storage.get_zone.return_value = unit.RoObject(
            name='example.com.',
            id=CentralZoneTestCase.zone_id,
            shared=False,
            tenant_id='t',
        )

        self.service.storage.create_zone_export = mock.Mock(
            return_value=unit.RwObject(
                id=CentralZoneTestCase.zone_export_id,
                zone_id=CentralZoneTestCase.zone_id,
                task_type='EXPORT',
                status='PENDING',
                message=None,
                tenant_id='t',
                location=None,
            )
        )

        self.service.worker_api.start_zone_export = mock.Mock()

        out = self.service.create_zone_export(
            self.context,
            CentralZoneTestCase.zone_export_id
        )
        self.assertEqual(CentralZoneTestCase.zone_id, out.zone_id)
        self.assertEqual('PENDING', out.status)
        self.assertEqual('EXPORT', out.task_type)
        self.assertIsNone(out.message)
        self.assertEqual('t', out.tenant_id)

    def test_get_zone_export(self):
        self.context = mock.Mock()
        self.context.project_id = 't'

        self.service.storage.get_zone_export.return_value = unit.RoObject(
            zone_id=CentralZoneTestCase.zone_id,
            task_type='EXPORT',
            status='PENDING',
            message=None,
            tenant_id='t'
        )

        out = self.service.get_zone_export(
            self.context,
            CentralZoneTestCase.zone_export_id)

        self.mock_policy_check.assert_called_with(
            'get_zone_export', mock.ANY, {'project_id': 't', 'tenant_id': 't'}
        )

        # Check output
        self.assertEqual(CentralZoneTestCase.zone_id, out.zone_id)
        self.assertEqual('PENDING', out.status)
        self.assertEqual('EXPORT', out.task_type)
        self.assertIsNone(out.message)
        self.assertEqual('t', out.tenant_id)

    def test_find_zone_exports(self):
        self.context = mock.Mock()
        self.context.project_id = 't'
        self.service.storage.find_zone_exports = mock.Mock()

        self.service.find_zone_exports(self.context)

        self.assertTrue(self.service.storage.find_zone_exports.called)

        self.mock_policy_check.assert_called_with(
            'find_zone_exports', mock.ANY, {
                'project_id': 't', 'tenant_id': 't'}
        )

    def test_find_zone_exports_with_custom_criterion(self):
        self.context = mock.Mock()
        self.context.project_id = 't'
        self.service.storage.find_zone_exports = mock.Mock()

        self.service.find_zone_exports(
            self.context, criterion={'project_id': 't'}
        )

        self.assertTrue(self.service.storage.find_zone_exports.called)

        self.mock_policy_check.assert_called_with(
            'find_zone_exports', mock.ANY, {
                'project_id': 't', 'tenant_id': 't'}
        )

    def test_delete_zone_export(self):
        self.context = mock.Mock()
        self.context.project_id = 't'

        self.service.storage.delete_zone_export = mock.Mock(
            return_value=unit.RoObject(
                zone_id=CentralZoneTestCase.zone_id,
                task_type='EXPORT',
                status='PENDING',
                message=None,
                tenant_id='t',
                type='PRIMARY',
            )
        )

        out = self.service.delete_zone_export(
            self.context,
            CentralZoneTestCase.zone_export_id)

        self.assertTrue(self.service.storage.delete_zone_export.called)

        self.assertEqual(CentralZoneTestCase.zone_id, out.zone_id)
        self.assertEqual('PENDING', out.status)
        self.assertEqual('EXPORT', out.task_type)
        self.assertIsNone(out.message)
        self.assertEqual('t', out.tenant_id)

        self.mock_policy_check.assert_called_with(
            'delete_zone_export', mock.ANY, {
                'zone_export_id': 'e887597f-9697-47dd-a202-7a2711f8669c',
                'project_id': 't', 'tenant_id': 't'
            }
        )


class CentralStatusTests(CentralBasic):
    def test_update_zone_or_record_status_no_zone(self):
        zone = unit.RwObject(
            id='uuid',
            action='CREATE',
            status='SUCCESS',
            serial=0,
        )

        self.service.storage.get_zone.return_value = zone
        self.service.storage.find_records.return_value = []

        new_zone = self.service.update_status(
            self.context, zone.id, 'NO_ZONE', 0, 'CREATE')

        self.assertEqual(new_zone.action, 'CREATE')
        self.assertEqual(new_zone.status, 'ERROR')

    def test_update_zone_or_record_status_handle_update_after_create(self):
        zone = unit.RwObject(
            id='uuid',
            action='UPDATE',
            status='PENDING',
            serial=0,
        )

        self.service.storage.get_zone.return_value = zone
        self.service.storage.find_records.return_value = []

        new_zone = self.service.update_status(
            self.context, zone.id, 'PENDING', 0, 'CREATE')

        self.assertEqual(new_zone.action, 'UPDATE')
        self.assertEqual(new_zone.status, 'PENDING')


class CentralQuotaTest(designate.tests.functional.TestCase):
    def setUp(self):
        super().setUp()
        self.config(quota_driver='noop')
        self.context = mock.Mock()
        self.zone = mock.Mock()
        self.quotas_of_one = {
            'zones': 1,
            'zone_recordsets': 1,
            'zone_records': 1,
            'recordset_records': 1,
            'api_export_size': 1
        }
        self.zone.shared = False
        self.service = service.Service()
        self.service._quota = None
        self.service._storage = mock.Mock()
        self.service.notifier = mock.Mock()
        self.service.quota.get_quotas = mock.Mock()

    def test_zone_record_quota_allows_lowering_value(self):
        self.service._quota = mock.Mock()
        self.service.storage.count_records.return_value = 10

        recordset = mock.Mock(spec=objects.RecordSet)
        recordset.managed = False
        recordset.records = ['1.1.1.%i' % (i + 1) for i in range(5)]

        self.service._enforce_record_quota(
            self.context, self.zone, recordset
        )

        # Ensure we check against the number of records that will
        # result in the API call. The 5 value is as if there were 10
        # unmanaged records unders a single recordset. We find 10
        # total - 10 for the recordset being passed in and add the 5
        # from the new recordset.
        check_zone_records = mock.call(
            self.context, self.zone.tenant_id, zone_records=10 - 10 + 5
        )
        self.assertIn(
            check_zone_records, self.service.quota.limit_check.mock_calls
        )

        # Check the recordset limit as well
        check_recordset_records = mock.call(
            self.context, self.zone.tenant_id, recordset_records=5
        )
        self.assertIn(
            check_recordset_records, self.service.quota.limit_check.mock_calls
        )

    def test_enforce_zone_quota(self):
        self.service.quota.get_quotas.return_value = self.quotas_of_one

        # Test creating one zone, 1 quota, no existing zones
        self.service.storage.count_zones.return_value = 0
        self.assertIsNone(
            self.service._enforce_zone_quota(self.context, 'fake_project_id')
        )

        # Test creating one zone, 1 quota, one existing zone
        self.service.storage.count_zones.return_value = 1
        self.assertRaisesRegex(
            exceptions.OverQuota,
            'Quota exceeded for zones',
            self.service._enforce_zone_quota,
            self.context, 'fake_project_id'
        )

    def test_enforce_recordset_quota(self):
        self.service.quota.get_quotas.return_value = self.quotas_of_one

        # Test creating one recordset, 1 quota, no existing recordsets
        self.service.storage.count_recordsets.return_value = 0
        self.assertIsNone(
            self.service._enforce_recordset_quota(
                self.context, self.zone
            )
        )

        # Test creating one recordset, 1 quota, one existing recordset
        self.service.storage.count_recordsets.return_value = 1
        self.assertRaisesRegex(
            exceptions.OverQuota,
            'Quota exceeded for zone_recordsets',
            self.service._enforce_recordset_quota,
            self.context, self.zone
        )

    def test_enforce_record_quota(self):
        self.service.quota.get_quotas.return_value = self.quotas_of_one

        self.service.storage.count_records.side_effect = [
            0, 0,
            1, 0,
            0, 1,
            1, 1,
            1, 1,
            1, 1,
        ]

        managed_recordset = mock.Mock(spec=objects.RecordSet)
        managed_recordset.managed = True

        recordset_one_record = mock.Mock()
        recordset_one_record.managed = False
        recordset_one_record.records = ['192.0.2.1']

        # Test that managed recordsets have no quota limit
        self.assertIsNone(
            self.service._enforce_record_quota(
                self.context, self.zone, managed_recordset
            )
        )
        self.service.storage.count_records.assert_not_called()

        # Test creating recordset with one record, no existing zone records,
        # no existing recordsets
        self.assertIsNone(
            self.service._enforce_record_quota(
                self.context, self.zone, recordset_one_record
            )
        )

        # Test creating recordset with one record, one existing zone record,
        # no exiting recordsets
        self.assertRaisesRegex(
            exceptions.OverQuota,
            'Quota exceeded for zone_records',
            self.service._enforce_record_quota, self.context, self.zone,
            recordset_one_record
        )

        # Test creating recordset with one record, one existing zone record,
        # no exiting recordsets
        # Note: Recordsets replace the existing recordset
        self.assertIsNone(
            self.service._enforce_record_quota(
                self.context, self.zone, recordset_one_record
            )
        )

        # Test creating recordset with one record, no existing zone record,
        # one exiting recordsets
        # Note: Recordsets replace the existing recordset
        self.assertIsNone(
            self.service._enforce_record_quota(
                self.context, self.zone, recordset_one_record
            )
        )

        recordset_two_record = mock.Mock()
        recordset_two_record.managed = False
        recordset_two_record.records = ['192.0.2.1', '192.0.2.2']

        # Test creating recordset with two records, one existing zone record,
        # one exiting recordsets
        self.assertRaisesRegex(
            exceptions.OverQuota,
            'Quota exceeded for zone_records',
            self.service._enforce_record_quota, self.context, self.zone,
            recordset_two_record
        )

        # Test creating a recordset with a shared zone
        mock_zone = MockZone()
        mock_zone.shared = True
        self.service.quota.limit_check = mock.Mock()
        self.service.storage.count_records = mock.Mock(return_value=1)
        self.service._enforce_record_quota(
            self.context, mock_zone, recordset_one_record
        )
        self.service.quota.limit_check.assert_called_with(
            self.context, mock_zone.tenant_id, recordset_records=1
        )



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate import objects
from designate.objects.adapters.api_v2 import base


class RecordAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.Record
    MODIFICATIONS = {
        'fields': {
            "data": {
                'read_only': False
            },
        },
        'options': {
            'links': False,
            'resource_name': 'record',
            'collection_name': 'records',
        }
    }

    @classmethod
    def render_object(cls, record, *arg, **kwargs):
        return record.data

    @classmethod
    def parse_object(cls, value, record_object, *args, **kwargs):
        record_object.data = value
        return record_object


class RecordListAPIv2Adapter(base.APIv2Adapter):
    ADAPTER_OBJECT = objects.RecordList
    MODIFICATIONS = {
        'options': {
            'links': False,
            'resource_name': 'record',
            'collection_name': 'records',
        }
    }

    @classmethod
    def render_list(cls, record_list, *arg, **kwargs):
        records = []
        for record in record_list:
            records.append(record.data)
        return records



# Copyright 2020 Cloudification GmbH. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from designate.objects import base
from designate.objects import fields


@base.DesignateRegistry.register
class SharedZone(base.DictObjectMixin, base.PersistentObjectMixin,
                 base.DesignateObject):
    fields = {
        'zone_id': fields.UUIDFields(nullable=False),
        'project_id': fields.StringFields(maxLength=36, nullable=False),
        'target_project_id': fields.StringFields(maxLength=36, nullable=False),
    }

    STRING_KEYS = [
        'id', 'zone_id', 'project_id', 'target_project_id'
    ]


@base.DesignateRegistry.register
class SharedZoneList(base.AttributeListObjectMixin, base.DesignateObject):
    LIST_ITEM_TYPE = SharedZone

    fields = {
        'objects': fields.ListOfObjectsField('SharedZone'),
    }
